# -*- encoding: utf-8 -*-
"""
@author: orange-crow, XuMing(xuming624@qq.com)
@description:
part of the code is from https://github.com/geekan/MetaGPT
"""
from __future__ import annotations

import base64
import re
import time
from typing import Literal, Tuple

try:
    from nbclient import NotebookClient
    from nbclient.exceptions import CellTimeoutError, DeadKernelError
    import nbformat
    from nbformat import NotebookNode
    from nbformat.v4 import new_code_cell, new_markdown_cell, new_output
except ImportError:
    raise ImportError("The `nbclient` package is not installed. Please install it via `pip install nbclient nbformat`.")

from rich.box import MINIMAL
from rich.console import Console, Group
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel
from rich.syntax import Syntax

from agentica.tools.base import Tool
from agentica.utils.log import logger


class RunNbCodeWrapper:
    """Run notebook code block, return result to llm, and display it."""

    nb: NotebookNode
    nb_client: NotebookClient
    console: Console
    interaction: str
    timeout: int = 600

    def __init__(
            self,
            nb=nbformat.v4.new_notebook(),
            timeout=600,
    ):
        self.nb = nb
        self.timeout = timeout
        self.console = Console()
        self.interaction = "ipython" if self.is_ipython() else "terminal"
        self.nb_client = NotebookClient(nb, timeout=timeout)

    def build(self):
        if self.nb_client.kc is None or not self.nb_client.kc.is_alive():
            self.nb_client.create_kernel_manager()
            self.nb_client.start_new_kernel()
            self.nb_client.start_new_kernel_client()

    def terminate(self):
        """Kill NotebookClient"""
        if self.nb_client.km is not None and self.nb_client.km.is_alive():
            self.nb_client.km.shutdown_kernel(now=True)
            self.nb_client.km.cleanup_resources()

            channels = [
                self.nb_client.kc.stdin_channel,
                self.nb_client.kc.hb_channel,
                self.nb_client.kc.control_channel,
            ]

            for channel in channels:
                if channel.is_alive():
                    channel.stop()

            self.nb_client.kc = None
            self.nb_client.km = None

    def reset(self):
        """Reset NotebookClient"""
        self.terminate()
        time.sleep(1)
        self.build()
        self.nb_client = NotebookClient(self.nb, timeout=self.timeout)

    def add_code_cell(self, code: str):
        self.nb.cells.append(new_code_cell(source=code))

    def add_markdown_cell(self, markdown: str):
        self.nb.cells.append(new_markdown_cell(source=markdown))

    def _display(self, code: str, language: Literal["python", "markdown"] = "python"):
        if language == "python":
            code = Syntax(code, "python", theme="paraiso-dark", line_numbers=True)
            self.console.print(code)
        elif language == "markdown":
            display_markdown(code)
        else:
            raise ValueError(f"Only support for python, markdown, but got {language}")

    def add_output_to_cell(self, cell: NotebookNode, output: str):
        """Add outputs of code execution to notebook cell."""
        if "outputs" not in cell:
            cell["outputs"] = []
        cell["outputs"].append(new_output(output_type="stream", name="stdout", text=str(output)))

    def parse_outputs(self, outputs: list, keep_len: int = 2000) -> Tuple[bool, str]:
        """Parses the outputs received from notebook execution."""
        assert isinstance(outputs, list)
        parsed_output, is_success = [], True
        for i, output in enumerate(outputs):
            output_text = ""
            if output["output_type"] == "stream" and not any(
                    tag in output["text"]
                    for tag in
                    ["| INFO     | agentica", "| ERROR    | agentica", "| WARNING  | agentica", "DEBUG"]
            ):
                output_text = output["text"]
            elif output["output_type"] == "display_data":
                if "image/png" in output["data"]:
                    self.show_bytes_figure(output["data"]["image/png"], self.interaction)
                else:
                    logger.info(
                        f"{i}th output['data'] from nbclient outputs dont have image/png, continue next output ..."
                    )
            elif output["output_type"] == "execute_result":
                output_text = output["data"]["text/plain"]
            elif output["output_type"] == "error":
                output_text, is_success = "\n".join(output["traceback"]), False

            # handle coroutines that are not executed asynchronously
            if output_text.strip().startswith("<coroutine object"):
                output_text = "Executed code failed, you need use key word 'await' to run a async code."
                is_success = False

            output_text = remove_escape_and_color_codes(output_text)
            # The useful information of the exception is at the end,
            # the useful information of normal output is at the begining.
            output_text = output_text[:keep_len] if is_success else output_text[-keep_len:]

            parsed_output.append(output_text)
        return is_success, ",".join(parsed_output)

    def show_bytes_figure(self, image_base64: str, interaction_type: str):
        image_bytes = base64.b64decode(image_base64)
        if interaction_type == "ipython":
            from IPython.display import Image, display

            display(Image(data=image_bytes))
        else:
            import io

            from PIL import Image

            image = Image.open(io.BytesIO(image_bytes))
            image.show()

    def is_ipython(self) -> bool:
        try:
            # 如果在Jupyter Notebook中运行，__file__ 变量不存在
            from IPython import get_ipython

            if get_ipython() is not None and "IPKernelApp" in get_ipython().config:
                return True
            else:
                return False
        except NameError:
            return False

    def run_cell(self, cell: NotebookNode, cell_index: int) -> Tuple[bool, str]:
        """Set timeout for run code."""
        try:
            self.nb_client.execute_cell(cell, cell_index)
            return self.parse_outputs(cell.outputs)
        except CellTimeoutError:
            assert self.nb_client.km is not None
            self.nb_client.km.interrupt_kernel()
            time.sleep(1)
            error_msg = "Cell execution timed out: Execution exceeded the time limit and was stopped; consider optimizing your code for better performance."
            return False, error_msg
        except DeadKernelError:
            self.reset()
            return False, "DeadKernelError"
        except Exception as e:
            return False, str(e)

    def run(self, code: str, language: Literal["python", "markdown"] = "python") -> Tuple[str, bool]:
        """
        Return the output of code execution, and a success indicator (bool) of code execution.
        """
        self._display(code, language)

        if language == "python":
            self.add_code_cell(code=code)
            self.build()
            cell_index = len(self.nb.cells) - 1
            success, outputs = self.run_cell(self.nb.cells[-1], cell_index)
            return outputs, success

        elif language == "markdown":
            # add markdown content to markdown cell in a notebook.
            self.add_markdown_cell(code)
            # return True, there is no execution failure for markdown cell.
            return code, True
        else:
            raise ValueError(f"Only support for language: python, markdown, but got {language}, ")


def remove_escape_and_color_codes(input_str: str):
    # 使用正则表达式去除jupyter notebook输出结果中的转义字符和颜色代码
    # Use regular expressions to get rid of escape characters and color codes in jupyter notebook output.
    pattern = re.compile(r"\x1b\[[0-9;]*[mK]")
    result = pattern.sub("", input_str)
    return result


def display_markdown(content: str):
    # Use regular expressions to match blocks of code one by one.
    matches = re.finditer(r"```(.+?)```", content, re.DOTALL)
    start_index = 0
    content_panels = []
    # Set the text background color and text color.
    style = "black on white"
    # Print the matching text and code one by one.
    for match in matches:
        text_content = content[start_index: match.start()].strip()
        code_content = match.group(0).strip()[3:-3]  # Remove triple backticks

        if text_content:
            content_panels.append(Panel(Markdown(text_content), style=style, box=MINIMAL))

        if code_content:
            content_panels.append(Panel(Markdown(f"```{code_content}"), style=style, box=MINIMAL))
        start_index = match.end()

    # Print remaining text (if any).
    remaining_text = content[start_index:].strip()
    if remaining_text:
        content_panels.append(Panel(Markdown(remaining_text), style=style, box=MINIMAL))

    # Display all panels in Live mode.
    with Live(auto_refresh=False, console=Console(), vertical_overflow="visible") as live:
        live.update(Group(*content_panels))
        live.refresh()


class RunNbCodeTool(Tool):
    """Run notebook code block, return result to llm, and display it."""

    def __init__(self, timeout: int = 600):
        super().__init__(name="run_nb_code_tool")
        self.executor = RunNbCodeWrapper(timeout=timeout)
        self.register(self.execute_nb_code)

    def execute_nb_code(self, code: str) -> str:
        """Execute a code block in a Jupyter Notebook and return the result.

        Args:
            code: str, The code to execute.

        Example:
            from agentica.tools.run_nb_code_tool import RunNbCodeTool
            m = RunNbCodeTool()
            r = m.execute_nb_code("import math\nprint(math.sqrt(79192201))")
            print(r)

        Returns:
            str, The result of the code execution.
        """
        try:
            outputs, success = self.executor.run(code)
            return outputs if success else "Execution failed"
        except Exception as e:
            return f"Execution failed with error: {str(e)}"


if __name__ == '__main__':
    executor = RunNbCodeWrapper()
    output, success = executor.run("print('Hello, world!')")
    print("Output:", output)
    print("Success:", success)

    # 添加并显示 Python 代码单元
    executor.add_code_cell("a = 10\nb = 20\nc = a + b\nprint(c)")
    executor._display("a = 10\nb = 20\nc = a + b\nprint(c)", "python")

    # 添加并显示 Markdown 单元
    executor.add_markdown_cell("# This is a markdown cell\n\nThis cell contains **bold** text and `inline code`.")
    executor._display("# This is a markdown cell\n\nThis cell contains **bold** text and `inline code`.", "markdown")

    # 构建 NotebookClient
    executor.build()

    # 运行第一个代码单元
    output, success = executor.run("a = 10\nb = 20\nc = a + b\nprint(c)", "python")
    print("Output:", output)
    print("Success:", success)

    # 运行第二个代码单元
    # 8899*8899=79192201
    output, success = executor.run("import math\nprint(math.sqrt(79192201))", "python")
    print("Output:", output)
    print("Success:", success)

    # 运行 Markdown 单元
    output, success = executor.run(
        "# This is a markdown cell\n\nThis cell contains **bold** text and `inline code`.",
        "markdown")
    print("Output:", output)
    print("Success:", success)

    # 终止 NotebookClient
    executor.terminate()

    # 重置 NotebookClient
    executor.reset()

    # 测试解析输出
    outputs = [
        {"output_type": "stream", "text": "This is a stream output\n"},
        {"output_type": "execute_result", "data": {"text/plain": "This is an execute result"}},
        {"output_type": "error", "traceback": ["Traceback (most recent call last):", "Error: Something went wrong"]}
    ]
    is_success, parsed_output = executor.parse_outputs(outputs)
    print("Parsed Output:", parsed_output)
    print("Is Success:", is_success)

    # 测试移除转义和颜色代码
    cleaned_output = remove_escape_and_color_codes("\x1b[31mThis is red text\x1b[0m")
    print("Cleaned Output:", cleaned_output)

    m = RunNbCodeTool()
    r = m.execute_nb_code("import math\nprint(math.sqrt(79192201))")
    print(type(r), '\n\n', r)



# -*- encoding: utf-8 -*-
"""
@author: orange-crow, XuMing(xuming624@qq.com)
@description:
part of the code is from phidata
"""
import json
import httpx
from agentica.tools.base import Tool
from agentica.utils.log import logger


class HackerNewsTool(Tool):
    def __init__(
        self,
        get_top_stories: bool = True,
        get_user_details: bool = True,
    ):
        super().__init__(name="hackers_news")

        # Register functions in the toolkit
        if get_top_stories:
            self.register(self.get_top_hackernews_stories)
        if get_user_details:
            self.register(self.get_user_details)

    def get_top_hackernews_stories(self, num_stories: int = 10) -> str:
        """Get top stories from Hacker News.

        Args:
            num_stories (int): Number of stories to return. Defaults to 10.

        Example:
            from agentica.tools.hackernews_tool import HackerNewsTool
            m = HackerNewsTool()
            top_stories = m.get_top_hackernews_stories(3)
            print(top_stories)

        Returns:
            str: JSON string of top stories.
        """

        logger.info(f"Getting top {num_stories} stories from Hacker News")
        # Fetch top story IDs
        response = httpx.get("https://hacker-news.firebaseio.com/v0/topstories.json")
        story_ids = response.json()

        # Fetch story details
        stories = []
        for story_id in story_ids[:num_stories]:
            story_response = httpx.get(f"https://hacker-news.firebaseio.com/v0/item/{story_id}.json")
            story = story_response.json()
            story["username"] = story["by"]
            stories.append(story)
        return json.dumps(stories, indent=2, ensure_ascii=False)

    def get_user_details(self, username: str) -> str:
        """Get the details of a Hacker News user using their username.

        Args:
            username (str): Username of the user to get details for.

        Example:
            from agentica.tools.hackernews_tool import HackerNewsTool
            m = HackerNewsTool()
            user_details = m.get_user_details('tomthe')
            print(user_details)

        Returns:
            str: JSON string of the user details.
        """

        try:
            logger.info(f"Getting details for user: {username}")
            user = httpx.get(f"https://hacker-news.firebaseio.com/v0/user/{username}.json").json()
            user_details = {
                "id": user.get("user_id"),
                "karma": user.get("karma"),
                "about": user.get("about"),
                "total_items_submitted": len(user.get("submitted", [])),
            }
            return json.dumps(user_details, indent=2, ensure_ascii=False)
        except Exception as e:
            logger.exception(e)
            return f"Error getting user details: {e}"


if __name__ == '__main__':
    m = HackerNewsTool()
    r = m.get_top_hackernews_stories(3)
    print(r)
    r = m.get_user_details('tomthe')
    print(r)


# -*- coding: utf-8 -*-
"""
@author: XuMing(xuming624@qq.com)
@description: Tools for editing files - useful for code editors and AI coding assistants.
"""
import os
import re
import difflib
from pathlib import Path
from typing import Optional, Dict, Any, List, Tuple

from agentica.tools.base import Tool
from agentica.utils.log import logger


class EditTool(Tool):
    """
    A toolkit for editing files, particularly useful for code editing features.
    Provides functionality similar to modern code editors.
    """

    def __init__(
            self,
            work_dir: Optional[str] = None,
            edit_file: bool = True,
            apply_patch: bool = True,
            compare_files: bool = True,
            search_replace: bool = True,
    ):
        """
        Initialize the EditTool.

        Args:
            work_dir: The working directory for file operations. Defaults to current directory.
            edit_file: Whether to include the edit_file function.
            apply_patch: Whether to include the apply_patch function.
            compare_files: Whether to include the compare_files function.
            search_replace: Whether to include the search_replace function.
        """
        super().__init__(name="edit_tool")
        self.work_dir = Path(work_dir) if work_dir else Path.cwd()
        
        if edit_file:
            self.register(self.edit_file)
        if apply_patch:
            self.register(self.apply_patch)
        if compare_files:
            self.register(self.compare_files)
        if search_replace:
            self.register(self.search_replace)

    def _resolve_path(self, file_path: str) -> Path:
        """Resolves a file path, making it absolute if it's relative."""
        path = Path(file_path)
        if not path.is_absolute():
            path = self.work_dir.joinpath(path)
        return path

    def edit_file(self, target_file: str, code_edit: str) -> str:
        """Edit a file or create a new one with the specified content.

        Args:
            target_file (str): The path of the file to edit.
            code_edit (str): The new content for the file or a patch in the format '*** Begin Patch\n...'

        Example:
            from agentica.tools.edit_tool import EditTool
            editor = EditTool()
            result = editor.edit_file("example.py", "print('Hello, World!')")
            print(result)

        Returns:
            str: A message describing the result of the operation.
        """
        try:
            file_path = self._resolve_path(target_file)
            
            # Check if this is a patch by looking for patch format markers
            if code_edit.startswith("*** Begin Patch") or "@@" in code_edit[:100]:
                return self.apply_patch(target_file, code_edit)
            
            # Ensure directory exists
            os.makedirs(file_path.parent, exist_ok=True)
            
            # Determine if we're creating or modifying
            action = "Created" if not file_path.exists() else "Updated"
            
            # Write the new content
            file_path.write_text(code_edit, encoding='utf-8')
            logger.info(f"{action} file: {file_path}")
            
            return f"{action} file: {target_file}"
            
        except Exception as e:
            error_msg = f"Error editing file {target_file}: {str(e)}"
            logger.error(error_msg)
            return error_msg

    def apply_patch(self, target_file: str, patch_content: str) -> str:
        """Apply a patch to a file.

        Args:
            target_file (str): The path of the file to patch.
            patch_content (str): The patch content in unified diff format.

        Example:
            from agentica.tools.edit_tool import EditTool
            editor = EditTool()
            patch = "*** Begin Patch ***\\n@@ -1,1 +1,2 @@\\n print('Hello')\\n+print('World')\\n"
            result = editor.apply_patch("example.py", patch)
            print(result)

        Returns:
            str: A message describing the result of the operation.
        """
        try:
            file_path = self._resolve_path(target_file)
            
            # Clean up the patch - remove any wrapper markers
            if "*** Begin Patch" in patch_content:
                # Extract just the diff part
                match = re.search(r'(?:\*{3} Begin Patch \*{3}[\r\n]+)([\s\S]+?)(?:[\r\n]+\*{3} End Patch \*{3}|$)', patch_content)
                if match:
                    patch_content = match.group(1)
            
            # Check if file exists
            if not file_path.exists():
                # For new files, just create with the content after the patch header
                # Extract the content after the patch header (anything after +++ line)
                content_match = re.search(r'\+{3}[^\n]*\n([\s\S]+)', patch_content)
                if content_match:
                    content = content_match.group(1)
                    # Remove the '+' markers at the beginning of lines
                    content = re.sub(r'^\+', '', content, flags=re.MULTILINE)
                    return self.edit_file(target_file, content)
                else:
                    return f"Error: Cannot apply patch to non-existent file {target_file}"
            
            # Read the original file
            original_content = file_path.read_text(encoding='utf-8')
            original_lines = original_content.splitlines()
            
            # Parse the patch into a list of diff objects
            patch_lines = patch_content.splitlines()
            
            # Apply the patch
            # This is a simplified patch application; for complex patches you might
            # want to use a proper diff/patch library
            try:
                patched_content = []
                i = 0
                while i < len(patch_lines):
                    line = patch_lines[i]
                    if line.startswith("@@"):
                        # Parse the hunk header
                        hunk_match = re.match(r'@@ -(\d+)(?:,\d+)? \+(\d+)(?:,\d+)? @@', line)
                        if hunk_match:
                            orig_start = int(hunk_match.group(1)) - 1  # 0-based indexing
                            new_start = int(hunk_match.group(2)) - 1   # 0-based indexing
                            
                            # Process the hunk
                            j = i + 1
                            orig_lines = []
                            new_lines = []
                            
                            while j < len(patch_lines) and not patch_lines[j].startswith("@@"):
                                hunk_line = patch_lines[j]
                                if hunk_line.startswith("+"):
                                    new_lines.append(hunk_line[1:])
                                elif hunk_line.startswith("-"):
                                    orig_lines.append(hunk_line[1:])
                                else:
                                    orig_lines.append(hunk_line[1:] if hunk_line.startswith(" ") else hunk_line)
                                    new_lines.append(hunk_line[1:] if hunk_line.startswith(" ") else hunk_line)
                                j += 1
                            
                            # Verify the original lines match
                            for k, orig_line in enumerate(orig_lines):
                                if orig_start + k >= len(original_lines) or original_lines[orig_start + k] != orig_line:
                                    raise ValueError(f"Patch does not match original file at line {orig_start + k + 1}")
                            
                            # Apply the changes
                            patched_content = original_lines[:orig_start] + new_lines + original_lines[orig_start + len(orig_lines):]
                            original_lines = patched_content
                            
                            i = j
                        else:
                            i += 1
                    else:
                        i += 1
                
                # Write the patched content
                file_path.write_text("\n".join(patched_content), encoding='utf-8')
                logger.info(f"Applied patch to file: {file_path}")
                
                return f"Successfully patched file: {target_file}"
                
            except ValueError as ve:
                return f"Error applying patch: {str(ve)}"
            
        except Exception as e:
            error_msg = f"Error applying patch to {target_file}: {str(e)}"
            logger.error(error_msg)
            return error_msg

    def compare_files(self, file1: str, file2: str, context_lines: int = 3) -> str:
        """Compare two files and return the differences.

        Args:
            file1 (str): The path to the first file.
            file2 (str): The path to the second file.
            context_lines (int): Number of context lines to show around differences.

        Example:
            from agentica.tools.edit_tool import EditTool
            editor = EditTool()
            result = editor.compare_files("file1.py", "file2.py")
            print(result)

        Returns:
            str: A unified diff of the two files.
        """
        try:
            path1 = self._resolve_path(file1)
            path2 = self._resolve_path(file2)
            
            if not path1.exists():
                return f"Error: File not found: {file1}"
            if not path2.exists():
                return f"Error: File not found: {file2}"
            
            # Read file contents
            content1 = path1.read_text(encoding='utf-8').splitlines()
            content2 = path2.read_text(encoding='utf-8').splitlines()
            
            # Generate unified diff
            diff = difflib.unified_diff(
                content1, content2, 
                fromfile=str(file1), tofile=str(file2),
                n=context_lines
            )
            
            # Convert the diff generator to a string
            diff_text = "\n".join(diff)
            
            return diff_text if diff_text else "Files are identical."
            
        except Exception as e:
            error_msg = f"Error comparing files {file1} and {file2}: {str(e)}"
            logger.error(error_msg)
            return error_msg

    def search_replace(self, target_file: str, search_pattern: str, replacement: str, use_regex: bool = False) -> str:
        """Search and replace text in a file.

        Args:
            target_file (str): The path of the file to modify.
            search_pattern (str): The text or pattern to search for.
            replacement (str): The text to replace with.
            use_regex (bool): Whether to treat search_pattern as a regular expression.

        Example:
            from agentica.tools.edit_tool import EditTool
            editor = EditTool()
            result = editor.search_replace("example.py", "Hello", "Hi")
            print(result)

        Returns:
            str: A message describing the result of the operation.
        """
        try:
            file_path = self._resolve_path(target_file)
            
            if not file_path.exists():
                return f"Error: File not found: {target_file}"
            
            # Read the file content
            content = file_path.read_text(encoding='utf-8')
            
            # Perform the replacement
            if use_regex:
                try:
                    pattern = re.compile(search_pattern, re.MULTILINE)
                    new_content, count = re.subn(pattern, replacement, content)
                except re.error as e:
                    return f"Error: Invalid regular expression pattern: {str(e)}"
            else:
                new_content, count = content.replace(search_pattern, replacement), content.count(search_pattern)
            
            # Only update the file if changes were made
            if count > 0:
                file_path.write_text(new_content, encoding='utf-8')
                logger.info(f"Replaced {count} occurrences in file: {file_path}")
                return f"Replaced {count} occurrences in file: {target_file}"
            else:
                return f"No occurrences of '{search_pattern}' found in {target_file}"
            
        except Exception as e:
            error_msg = f"Error performing search and replace in {target_file}: {str(e)}"
            logger.error(error_msg)
            return error_msg


if __name__ == '__main__':
    # Simple test
    editor = EditTool()
    
    # Test edit_file
    print(editor.edit_file("test_edit.py", "print('Hello, World!')"))
    
    # Test search_replace
    print(editor.search_replace("test_edit.py", "Hello", "Greetings"))
    
    # Clean up
    import os
    if os.path.exists("test_edit.py"):
        os.remove("test_edit.py") 


# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
import os
from datetime import datetime
from dotenv import load_dotenv  # noqa
from loguru import logger  # noqa, need to import logger here to avoid circular import

AGENTICA_HOME = os.getenv("AGENTICA_HOME", os.path.expanduser("~/.agentica"))
if AGENTICA_HOME:
    os.makedirs(AGENTICA_HOME, exist_ok=True)

# Load environment variables from .env file
AGENTICA_DOTENV_PATH = os.getenv("AGENTICA_DOTENV_PATH", f"{AGENTICA_HOME}/.env")
load_dotenv(AGENTICA_DOTENV_PATH, override=True)

AGENTICA_DATA_DIR = os.getenv("AGENTICA_DATA_DIR", f"{AGENTICA_HOME}/data")
AGENTICA_LOG_LEVEL = os.getenv("AGENTICA_LOG_LEVEL", "INFO").upper()
AGENTICA_LOG_FILE = os.getenv("AGENTICA_LOG_FILE")
if AGENTICA_LOG_LEVEL == "DEBUG":
    formatted_date = datetime.now().strftime("%Y%m%d")
    default_log_file = f"{AGENTICA_HOME}/logs/{formatted_date}.log"
    AGENTICA_LOG_FILE = os.getenv("AGENTICA_LOG_FILE", default_log_file)
    logger.debug(f"AGENTICA_LOG_LEVEL: DEBUG, AGENTICA_LOG_FILE: {AGENTICA_LOG_FILE}")
    if os.path.exists(AGENTICA_DOTENV_PATH):
        logger.debug(f"Loaded AGENTICA_DOTENV_PATH: {AGENTICA_DOTENV_PATH}")



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: Model Context Protocol (MCP) client implementations supporting both stdio and SSE transports
"""
import asyncio
import concurrent.futures
import copy
import json
import threading
from os import environ
from typing import Dict, Optional, Union
from datetime import timedelta

from mcp import ClientSession, StdioServerParameters
from mcp.client.sse import sse_client
from mcp.client.stdio import stdio_client
from mcp.client.streamable_http import streamablehttp_client

from agentica.mcp.client import MCPClient
from agentica.mcp.config import MCPConfig
from agentica.mcp.server import MCPServerStdio, MCPServerSse, MCPServerStreamableHttp
from agentica.tools.base import Function
from agentica.tools.base import Tool
from agentica.utils.log import logger


class McpTool(Tool):
    """
    A toolkit for integrating Model Context Protocol (MCP) servers with Agno agents.
    This allows agents to access tools, resources, and prompts exposed by MCP servers.

    Can be used in three ways:
    1. Direct initialization with a ClientSession
    2. As an async context manager with StdioServerParameters or stdio_command for stdio transport
    3. As an async context manager with sse_server_url for SSE transport
    """

    def __init__(
            self,
            command: Optional[str] = None,
            url: Optional[str] = None,
            *,
            env: Optional[dict[str, str]] = None,
            server_params: Optional[StdioServerParameters] = None,
            session: Optional[ClientSession] = None,
            include_tools: Optional[list[str]] = None,
            exclude_tools: Optional[list[str]] = None,
            sse_headers: Optional[Dict[str, str]] = None,
            sse_timeout: float = 5.0,
            sse_read_timeout: float = 300.0,
            terminate_on_close: bool = True,
    ):
        """
        Initialize the MCP toolkit.

        Args:
            command: The command to run to start the stdio server. Should be used in conjunction with env.
            url: The URL of the SSE or StreamableHttp endpoint.
            env: The environment variables to pass to the server. Used with stdio_command or for additional SSE config.
            server_params: StdioServerParameters for creating a new stdio session
            session: An initialized MCP ClientSession connected to an MCP server
            include_tools: Optional list of tool names to include (if None, includes all)
            exclude_tools: Optional list of tool names to exclude (if None, excludes none)
            sse_headers: Optional headers for the SSE or StreamableHttp connection
            sse_timeout: HTTP request timeout for SSE/StreamableHttp (default: 5 seconds)
            sse_read_timeout: SSE/StreamableHttp connection timeout (default: 300 seconds)
            terminate_on_close: Whether to terminate on close (StreamableHttp only, default: True)
        """
        super().__init__(name="McpTool")

        # Validate that at least one connection method is provided
        if session is None and server_params is None and command is None and url is None and env is None:
            raise ValueError("Either session, server_params, stdio_command, or sse_server_url must be provided")

        self.session: Optional[ClientSession] = session
        self.server_params: Optional[StdioServerParameters] = server_params

        # Merge provided env with system env
        if env is not None:
            env = {
                **environ,
                **env,
            }
        else:
            env = {**environ}

        # Determine the transport type (stdio or SSE)
        self._transport_type = "stdio"  # Default to stdio

        # Check for direct URL parameter first
        self._url = url

        # If not provided directly, check environment
        if not self._url:
            self._url = env.get("MCP_SERVER_URL")

        # Configure SSE or StreamableHttp if URL is available
        if self._url:
            # Determine transport type based on URL
            # URLs with "/sse" are treated as SSE
            # All other URLs are treated as StreamableHttp
            if "/sse" in self._url:
                self._transport_type = "sse"
            else:
                # Default to streamable-http for all other URLs
                self._transport_type = "streamable-http"

            # Headers provided directly take precedence
            if sse_headers is not None:
                self._headers = sse_headers
            else:
                # Try to get headers from environment
                env_headers = env.get("MCP_SERVER_HEADERS", {})
                if isinstance(env_headers, str):
                    try:
                        self._headers = json.loads(env_headers)
                    except:
                        self._headers = {}
                else:
                    self._headers = env_headers

            # Timeouts provided directly take precedence
            self._timeout = sse_timeout
            self._read_timeout = sse_read_timeout
            self._terminate_on_close = terminate_on_close

            # If not provided directly, try to get from environment
            if "MCP_SERVER_TIMEOUT" in env:
                try:
                    self._timeout = float(env.get("MCP_SERVER_TIMEOUT", "5"))
                except ValueError:
                    self._timeout = 5.0

            if "MCP_SERVER_READ_TIMEOUT" in env:
                try:
                    self._read_timeout = float(env.get("MCP_SERVER_READ_TIMEOUT", "300"))
                except ValueError:
                    self._read_timeout = 300.0

        # Configure stdio if command is provided and we're not using SSE or StreamableHttp
        elif command is not None:
            from shlex import split

            parts = split(command)
            if not parts:
                raise ValueError("Empty command string")
            cmd = parts[0]
            arguments = parts[1:] if len(parts) > 1 else []
            self.server_params = StdioServerParameters(command=cmd, args=arguments, env=env)

            # Store the original command for easier reuse and debugging
            self._stdio_command = command
            self._cmd_parts = parts

        self._transport_context = None
        self._session_context = None
        self._initialized = False

        self.include_tools = include_tools
        self.exclude_tools = exclude_tools or []

        # Store the server configuration separately for tool functions to use directly
        self._server_config = {
            "transport_type": self._transport_type,
            "command": self.server_params.command if self.server_params and self._transport_type == "stdio" else None,
            "args": self.server_params.args if self.server_params and self._transport_type == "stdio" else [],
            "env": env,
            "url": self._url if self._transport_type in ["sse", "streamable-http"] else None,
            "headers": self._headers if self._transport_type in ["sse", "streamable-http"] else None,
            "timeout": self._timeout if self._transport_type in ["sse", "streamable-http"] else None,
            "read_timeout": self._read_timeout if self._transport_type in ["sse", "streamable-http"] else None,
            "terminate_on_close": self._terminate_on_close if self._transport_type == "streamable-http" else None,
        }

    async def __aenter__(self) -> "McpTool":
        """Enter the async context manager."""

        if self.session is not None:
            # Already has a session, just initialize
            if not self._initialized:
                await self.initialize()
            return self

        try:
            # Handle different transport types
            if self._transport_type == "sse":
                # Create an SSE client connection
                if not self._url:
                    raise ValueError("url or MCP_SERVER_URL must be provided for SSE transport")

                self._transport_context = sse_client(
                    url=self._url,
                    headers=self._headers,
                    timeout=self._timeout,
                    sse_read_timeout=self._read_timeout
                )
                
                # For SSE, we get a tuple of (read, write)
                read, write = await self._transport_context.__aenter__()
                
            elif self._transport_type == "streamable-http":
                # Create a StreamableHttp client connection
                if not self._url:
                    raise ValueError("url or MCP_SERVER_URL must be provided for StreamableHttp transport")

                self._transport_context = streamablehttp_client(
                    url=self._url,
                    headers=self._headers,
                    timeout=timedelta(seconds=self._timeout),
                    sse_read_timeout=timedelta(seconds=self._read_timeout),
                    terminate_on_close=self._terminate_on_close
                )
                
                # For StreamableHttp, we get a tuple of (read, write, get_session_id)
                transport_result = await self._transport_context.__aenter__()
                read, write = transport_result[0], transport_result[1]
                
            else:
                # Create a stdio client connection
                if self.server_params is None:
                    raise ValueError("server_params or stdio_command must be provided when using stdio transport")

                self._transport_context = stdio_client(self.server_params)
                
                # For stdio, we get a tuple of (read, write)
                read, write = await self._transport_context.__aenter__()

            # Create session from transport
            self._session_context = ClientSession(read, write)
            self.session = await self._session_context.__aenter__()

            # Store the parameters for later use in tool functions
            if self._transport_type == "stdio" and not hasattr(self.session, "_server_params"):
                setattr(self.session, "_server_params", self.server_params)
            elif self._transport_type in ["sse", "streamable-http"] and not hasattr(self.session, "_url"):
                setattr(self.session, "_url", self._url)

            # Initialize with the new session
            await self.initialize()
            return self
        except Exception as e:
            logger.error(f"Failed to enter MCP tool context: {e}")
            # Try to clean up any partially initialized contexts
            if self._session_context is not None:
                try:
                    await self._session_context.__aexit__(None, None, None)
                except:
                    pass
            if self._transport_context is not None:
                try:
                    await self._transport_context.__aexit__(None, None, None)
                except:
                    pass
            raise

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exit the async context manager."""
        try:
            if self._session_context is not None:
                await self._session_context.__aexit__(exc_type, exc_val, exc_tb)
                self.session = None
                self._session_context = None

            if self._transport_context is not None:
                try:
                    await self._transport_context.__aexit__(exc_type, exc_val, exc_tb)
                except GeneratorExit:
                    # This is expected for streamable-http transport during cleanup
                    if self._transport_type == "streamable-http":
                        logger.debug("GeneratorExit during streamable-http transport cleanup (expected)")
                    else:
                        logger.warning("GeneratorExit during transport cleanup")
                except Exception as e:
                    logger.error(f"Error closing transport context: {e}")
                self._transport_context = None
        except Exception as e:
            logger.error(f"Error exiting MCP tool context: {e}")
        finally:
            self._initialized = False

    async def initialize(self) -> None:
        """Initialize the MCP toolkit by getting available tools from the MCP server"""
        if self._initialized:
            return

        try:
            if self.session is None:
                raise ValueError("Session is not available. Use as context manager or provide a session.")

            # Initialize the session if not already initialized
            try:
                await asyncio.wait_for(self.session.initialize(), timeout=10.0)
            except Exception as e:
                logger.error(f"Error initializing MCP session: {e}")
                raise ValueError(f"Error initializing MCP session: {e}")

            # Get the list of tools from the MCP server
            try:
                available_tools = await asyncio.wait_for(self.session.list_tools(), timeout=10.0)
                logger.debug(f"Available MCP tools: {[t.name for t in available_tools.tools]}")
            except Exception as e:
                logger.error(f"Error listing MCP tools: {e}")
                raise ValueError(f"Error listing MCP tools: {e}")

            # Filter tools based on include/exclude lists
            filtered_tools = []
            for tool in available_tools.tools:
                if tool.name in self.exclude_tools:
                    continue
                if self.include_tools is None or tool.name in self.include_tools:
                    filtered_tools.append(tool)

            # Register the tools with the toolkit
            for tool in filtered_tools:
                try:
                    # Create a custom entrypoint that matches the transport type pattern
                    tool_name = tool.name

                    # Define a wrapper function that creates a fresh connection each time
                    def create_tool_function(t_name):
                        """Create a function for the tool that creates a fresh connection each time"""

                        async def _call_mcp_tool_async(**kwargs):
                            """Async implementation that creates a fresh connection"""
                            logger.debug(f"Creating new MCP connection for tool '{t_name}'")

                            # Create appropriate server based on transport type
                            if self._server_config["transport_type"] == "sse":
                                server = MCPServerSse(
                                    name=t_name,
                                    params={
                                        "url": self._server_config["url"],
                                        "headers": self._server_config["headers"],
                                        "timeout": self._server_config["timeout"],
                                        "sse_read_timeout": self._server_config["read_timeout"]
                                    }
                                )
                            elif self._server_config["transport_type"] == "streamable-http":
                                server = MCPServerStreamableHttp(
                                    name=t_name,
                                    params={
                                        "url": self._server_config["url"],
                                        "headers": self._server_config["headers"],
                                        "timeout": timedelta(seconds=self._server_config["timeout"]) if self._server_config["timeout"] else None,
                                        "sse_read_timeout": timedelta(seconds=self._server_config["read_timeout"]) if self._server_config["read_timeout"] else None,
                                        "terminate_on_close": self._server_config["terminate_on_close"]
                                    }
                                )
                            else:
                                server = MCPServerStdio(
                                    name=t_name,
                                    params={
                                        "command": self._server_config["command"],
                                        "args": self._server_config["args"],
                                        "env": copy.deepcopy(self._server_config["env"])
                                    }
                                )

                            try:
                                async with MCPClient(server=server) as client:
                                    # Call the tool (with timeout)
                                    read_timeout = self._server_config["read_timeout"] or 300.0
                                    result = await asyncio.wait_for(
                                        client.call_tool(t_name, kwargs),
                                        timeout=read_timeout
                                    )
                                    # Check if result is a string (error message) or a CallToolResult
                                    if isinstance(result, str):
                                        return result  # Already an error message string
                                    return client.extract_result_text(result)
                            except Exception as e:
                                logger.error(f"Error calling MCP tool '{t_name}': {e}")
                                return f"Error calling MCP tool '{t_name}': {e}"

                        def tool_function(**kwargs):
                            """Synchronous wrapper that handles async execution safely"""
                            future = concurrent.futures.Future()

                            def run_in_thread():
                                try:
                                    # Create new event loop
                                    loop = asyncio.new_event_loop()
                                    asyncio.set_event_loop(loop)
                                    try:
                                        # Run the async function
                                        result = loop.run_until_complete(_call_mcp_tool_async(**kwargs))
                                        future.set_result(result)
                                    finally:
                                        loop.close()
                                except Exception as e:
                                    future.set_exception(e)

                            # Start thread and wait for result
                            thread = threading.Thread(target=run_in_thread)
                            thread.start()
                            thread.join(timeout=60)

                            if thread.is_alive():
                                return f"Error: Timeout calling MCP tool '{t_name}'"

                            return future.result()

                        tool_function.__name__ = f"mcp_tool_{t_name}"
                        tool_function.__doc__ = f"Call the MCP tool '{t_name}'"

                        return tool_function

                    # Create the function for this tool
                    entrypoint = create_tool_function(tool_name)

                    # Create parameter schema
                    tool_params = {"type": "object", "properties": {}}
                    if hasattr(tool, 'inputSchema') and tool.inputSchema:
                        if isinstance(tool.inputSchema, dict):
                            if "properties" in tool.inputSchema:
                                tool_params = tool.inputSchema
                            else:
                                tool_params = {
                                    "type": "object",
                                    "properties": tool.inputSchema
                                }

                    f = Function(
                        name=tool_name,
                        description=tool.description,
                        parameters=tool_params,
                        entrypoint=entrypoint,
                        sanitize_arguments=False,  # We'll handle this in the entrypoint
                        skip_entrypoint_processing=True  # Skip processing to preserve our custom entrypoint
                    )
                    self.functions[f.name] = f
                    logger.debug(f"Function: {f.name} registered with {self.name}, Function parameters: {tool_params}")
                except Exception as e:
                    logger.error(f"Failed to register tool {tool.name}: {e}")

            logger.debug(
                f"{self.name} initialized with {len(filtered_tools)} tools using {self._transport_type} transport")
            self._initialized = True
        except Exception as e:
            logger.error(f"Failed to get MCP tools: {e}")
            raise

    @classmethod
    def from_config(cls, server_names: Optional[Union[str, list[str]]] = None,
                    config_path: Optional[str] = None):
        """Create McpTool instance from configuration file.

        Args:
            server_names: Optional server name(s) to load. If None, loads all servers
            config_path: Optional path to config file

        Returns:
            McpTool instance with configured servers

        Raises:
            ValueError: If specified servers not found in config
        """
        config = MCPConfig(config_path)
        all_servers = config.servers

        if not all_servers:
            raise ValueError(f"No MCP servers found in configuration: `{config.config_path}`")

        # Convert single server name to list
        if isinstance(server_names, str):
            server_names = [server_names]

        # If no servers specified, use all
        if server_names is None:
            server_names = list(all_servers.keys())

        # Validate servers exist
        invalid_servers = set(server_names) - set(all_servers.keys())
        if invalid_servers:
            raise ValueError(f"Servers not found in MCP configuration: {invalid_servers}. "
                             f"Available servers: {list(all_servers.keys())}")

        tools = []
        for name in server_names:
            server = all_servers[name]
            if server.url:
                # Determine transport type based on URL
                # URLs with "/sse" are treated as SSE
                # All other URLs are treated as StreamableHttp
                is_streamable_http = True  # Default to streamable-http
                transport_type = "streamable-http"
                
                if "/sse" in server.url:
                    is_streamable_http = False
                    transport_type = "sse"
                
                # Create tool with appropriate configuration
                tool = cls(
                    url=server.url,
                    sse_headers=server.headers,
                    sse_timeout=server.timeout,
                    sse_read_timeout=server.read_timeout,
                    terminate_on_close=True if is_streamable_http else None
                )
                
                logger.debug(f"Created {transport_type} tool for server '{name}' with URL: {server.url}")
            else:
                # Create stdio tool
                cmd = f"{server.command} {' '.join(server.args or [])}".strip()
                tool = cls(
                    command=cmd,
                    env=server.env
                )
                logger.debug(f"Created stdio tool for server '{name}' with command: {cmd}")
            tools.append(tool)

        logger.info(f"Created {len(tools)} MCP tools({server_names}) from configuration: {config.config_path}")

        # Return single tool or composite
        if len(tools) == 1:
            return tools[0]
        return CompositeMultiMcpTool(tools)


class CompositeMultiMcpTool(McpTool):
    """Combines multiple McpTool instances into one."""

    def __init__(self, tools: list[McpTool]):
        super().__init__(command="dummy")  # Dummy command to satisfy parent init
        self.tools = tools
        self.functions = {}
        self._initialized = False

    async def _enter_tool(self, tool):
        """Helper method to enter a single tool context."""
        try:
            entered_tool = await tool.__aenter__()
            await entered_tool.initialize()
            return entered_tool
        except Exception as e:
            logger.error(f"Error entering tool context: {e}")
            # Don't re-raise, just return None so we can continue with other tools
            return None

    async def __aenter__(self):
        """Enter context for all tools sequentially."""
        entered_tools = []
        errors = []
        
        for tool in self.tools:
            try:
                entered_tool = await self._enter_tool(tool)
                if entered_tool:
                    entered_tools.append(entered_tool)
            except Exception as e:
                errors.append((tool, e))
                logger.error(f"Failed to initialize tool: {e}")
        
        self.tools = [t for t in entered_tools if t is not None]
        
        # Merge functions from all successfully initialized tools
        for tool in self.tools:
            self.functions.update(tool.functions)

        self._initialized = True
        
        if not self.tools:
            if errors:
                error_msgs = "\n".join([f"- {e}" for _, e in errors])
                raise ValueError(f"Failed to initialize any tools. Errors:\n{error_msgs}")
            else:
                raise ValueError("No tools were successfully initialized")
                
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exit context for all tools sequentially."""
        errors = []
        for tool in reversed(self.tools):  # Exit in reverse order
            try:
                await tool.__aexit__(exc_type, exc_val, exc_tb)
            except Exception as e:
                errors.append(e)
                logger.error(f"Error exiting tool context: {e}")

        if errors and not exc_type:  # Only raise if we're not already handling an exception
            raise errors[0]

    async def initialize(self):
        """Initialize is handled during __aenter__."""
        if self._initialized:
            return



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: Use EasyOCR to extract text from images.
"""
try:
    from imgocr import ImgOcr
except ImportError:
    raise ImportError("The `imgocr` package is not installed. Please install it via `pip install imgocr`.")

from agentica.tools.base import Tool
from agentica.utils.log import logger


class OcrTool(Tool):
    def __init__(self, use_gpu: bool = False):
        """
        Initializes the OCR tool with the specified languages and GPU setting.

        :param languages: List of languages to use for OCR.
        :param use_gpu: Whether to use GPU for OCR processing.
        """
        super().__init__(name="ocr_tool")
        self.ocr_model = ImgOcr(use_gpu=use_gpu)
        logger.debug(f"Initialized imgocr tool, use GPU: {use_gpu}")
        self.register(self.read_text)

    def read_text(self, image_path: str) -> str:
        """Reads text from an image.

        Args:
            image_path (str): Path to the image file.

        Example:
            from agentica.tools.ocr_tool import OcrTool
            ocr_tool = OcrTool()
            text = ocr_tool.read_text('../../examples/data/chinese.jpg')
            print('Text:', text)

        Returns:
            str: The recognized text.
        """
        try:
            result = self.ocr_model.ocr(image_path)
            result_str = " ".join([i['text'] for i in result])
            logger.info(f"Ocr image: {image_path}, result: {result_str}")
            return result_str
        except Exception as e:
            logger.error(f"An error occurred while reading the text: {e}")
            return ""


if __name__ == '__main__':
    # Initialize the OCR tool
    ocr_tool = OcrTool()
    chinese_text = ocr_tool.read_text('../../examples/data/chinese.jpg')
    print('Chinese text:', chinese_text)



# -*- encoding: utf-8 -*-
"""
@author: orange-crow, XuMing(xuming624@qq.com)
@description:
part of the code is from phidata
"""
import json
import ssl
from typing import Any, Optional

from agentica.tools.base import Tool
from agentica.utils.log import logger

try:
    from duckduckgo_search import DDGS
except ImportError:
    raise ImportError("`duckduckgo-search` not installed. Please install using `pip install duckduckgo-search`")

# Create a default context for HTTPS requests (not recommended for production)
ssl._create_default_https_context = ssl._create_unverified_context


class DuckDuckGoTool(Tool):
    def __init__(
            self,
            search: bool = True,
            news: bool = True,
            headers: Optional[Any] = None,
            proxy: Optional[Any] = None,
            timeout: Optional[int] = 10,
    ):
        super().__init__(name="duckduckgo_tool")

        self.headers: Optional[Any] = headers
        self.proxy: Optional[Any] = proxy
        self.timeout: Optional[int] = timeout
        self.ddgs = DDGS(headers=self.headers, proxies=self.proxy, timeout=self.timeout)
        if search:
            self.register(self.duckduckgo_search)
        if news:
            self.register(self.duckduckgo_news)

    @staticmethod
    def search_with_ddgs(query: str):
        """
        Search with ddgs and return the contexts.
        """
        contexts = []
        search_results = []
        with DDGS() as ddgs:
            ddgs_gen = ddgs.text(query, backend="lite", timelimit="d, w, m, y")
            for r in ddgs_gen:
                search_results.append(r)
        for idx, result in enumerate(search_results):
            if result["body"] and result["href"]:
                contexts.append({
                    "name": result["title"],
                    "url": result["href"],
                    "snippet": result["body"]
                })
        return contexts

    def duckduckgo_search(self, query: str, max_results: int = 5) -> str:
        """Search DuckDuckGo for a query.

        Args:
            query(str): The query to search for.
            max_results (optional, default=5): The maximum number of results to return.

        Example:
            from agentica.tools.duckduckgo_tool import DuckDuckGoTool
            m = DuckDuckGoTool()
            search_results = m.duckduckgo_search("Python")
            print(search_results)

        Returns:
            The result from DuckDuckGo, in JSON format. The result includes the title, URL, and snippet.
        """
        logger.debug(f"Searching DDG for: {query}")
        gen_res = self.ddgs.text(query, backend="lite", timelimit="d, w, m, y")
        res = list(gen_res)[:max_results]
        return json.dumps(res, indent=2, ensure_ascii=False)

    def duckduckgo_news(self, query: str, max_results: int = 5) -> str:
        """Get the latest news from DuckDuckGo.

        Args:
            query(str): The query to search for.
            max_results (optional, default=5): The maximum number of results to return.

        Example:
            from agentica.tools.duckduckgo_tool import DuckDuckGoTool
            m = DuckDuckGoTool()
            news_results = m.duckduckgo_news("Python")
            print(news_results)

        Returns:
            The latest news from DuckDuckGo.
        """
        logger.debug(f"Searching DDG news for: {query}")
        gen_res = self.ddgs.news(query, timelimit="d, w, m, y")
        res = list(gen_res)[:max_results]
        return json.dumps(res, indent=2, ensure_ascii=False)


if __name__ == '__main__':
    # from agentica.tools.duckduckgo_tool import DuckDuckGoTool
    m = DuckDuckGoTool()
    print(m.duckduckgo_search("Python"))
    print(m.duckduckgo_news("Python"))



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
part of the code from https://github.com/phidatahq/phidata
"""
from typing import List, Optional, Callable, Any

from agentica.document import Document
from agentica.knowledge.base import Knowledge
from agentica.utils.log import logger


class LangChainKnowledge(Knowledge):
    loader: Optional[Callable] = None

    vectorstore: Optional[Any] = None
    search_kwargs: Optional[dict] = None

    retriever: Optional[Any] = None

    def search(self, query: str, num_documents: Optional[int] = None, filters=None) -> List[Document]:
        """Returns relevant documents matching the query"""

        try:
            from langchain_core.vectorstores import VectorStoreRetriever
            from langchain_core.documents import Document as LangChainDocument
        except ImportError:
            raise ImportError(
                "The `langchain` package is not installed. Please install it via `pip install langchain`."
            )

        if self.vectorstore is not None and self.retriever is None:
            logger.debug("Creating retriever")
            if self.search_kwargs is None:
                self.search_kwargs = {"k": self.num_documents}
            self.retriever = self.vectorstore.as_retriever(search_kwargs=self.search_kwargs)

        if self.retriever is None:
            logger.error("No retriever provided")
            return []

        if not isinstance(self.retriever, VectorStoreRetriever):
            raise ValueError(f"Retriever is not of type VectorStoreRetriever: {self.retriever}")

        _num_documents = num_documents or self.num_documents
        logger.debug(f"Getting {_num_documents} relevant documents for query: {query}")
        lc_documents: List[LangChainDocument] = self.retriever.invoke(input=query)
        documents = []
        for lc_doc in lc_documents:
            documents.append(
                Document(
                    content=lc_doc.page_content,
                    meta_data=lc_doc.metadata,
                )
            )
        return documents

    def load(self, recreate: bool = False, upsert: bool = True, skip_existing: bool = True, filters=None) -> None:
        if self.loader is None:
            logger.error("No loader provided for LangChainKnowledgeBase")
            return
        self.loader()

    def exists(self) -> bool:
        logger.warning("LangChainKnowledgeBase.exists() not supported - please check the vectorstore manually.")
        return True



import json
from dataclasses import dataclass, field
from typing import Optional, List, Iterator, Dict, Any, Union, Callable

from agentica.model.base import Model
from agentica.model.message import Message
from agentica.model.response import ModelResponse
from agentica.tools.base import Function, FunctionCall, ModelTool, Tool, get_function_call_for_tool_call
from agentica.utils.log import logger
from agentica.utils.timer import Timer

try:
    from vertexai.generative_models import (
        GenerativeModel,
        GenerationResponse,
        FunctionDeclaration,
        Tool as GeminiTool,
        Candidate,
        Content,
        Part,
    )
except (ModuleNotFoundError, ImportError):
    raise ImportError(
        "`google-cloud-aiplatform` not installed. Please install using `pip install google-cloud-aiplatform`"
    )


@dataclass
class MessageData:
    response_content: str = ""
    response_block: Content = None
    response_candidates: Optional[List[Candidate]] = None
    response_role: Optional[str] = None
    response_parts: Optional[List] = None
    response_tool_calls: List[Dict[str, Any]] = field(default_factory=list)
    response_usage: Optional[Dict[str, Any]] = None
    response_tool_call_block: Content = None


@dataclass
class Metrics:
    input_tokens: int = 0
    output_tokens: int = 0
    total_tokens: int = 0
    time_to_first_token: Optional[float] = None
    response_timer: Timer = field(default_factory=Timer)

    def log(self):
        if self.time_to_first_token is not None:
            logger.debug(f"* Time to first token:         {self.time_to_first_token:.4f}s")
        logger.debug(f"* Time to generate response:   {self.response_timer.elapsed:.4f}s")
        logger.debug(f"* Tokens per second:           {self.output_tokens / self.response_timer.elapsed:.4f} tokens/s")
        logger.debug(f"* Input tokens:                {self.input_tokens}")
        logger.debug(f"* Output tokens:               {self.output_tokens}")
        logger.debug(f"* Total tokens:                {self.total_tokens}")


class Gemini(Model):
    """

    Class for interacting with the VertexAI Gemini API.

    Attributes:

        name (str): The name of the API. Default is "Gemini".
        model (str): The model name. Default is "gemini-1.5-flash-002".
        provider (str): The provider of the API. Default is "VertexAI".
        generation_config (Optional[Any]): The generation configuration.
        safety_settings (Optional[Any]): The safety settings.
        generative_model_request_params (Optional[Dict[str, Any]]): The generative model request parameters.
        function_declarations (Optional[List[FunctionDeclaration]]): The function declarations.
        client (Optional[GenerativeModel]): The GenerativeModel client.
    """

    id: str = "gemini-2.0-flash-exp"
    name: str = "Gemini"
    provider: str = "VertexAI"

    # Request parameters
    generation_config: Optional[Any] = None
    safety_settings: Optional[Any] = None
    generative_model_request_params: Optional[Dict[str, Any]] = None
    function_declarations: Optional[List[FunctionDeclaration]] = None

    # Gemini client
    client: Optional[GenerativeModel] = None

    def get_client(self) -> GenerativeModel:
        """
        Returns a GenerativeModel client.

        Returns:
            GenerativeModel: GenerativeModel client.
        """
        if self.client is None:
            self.client = GenerativeModel(model_name=self.id, **self.request_kwargs)
        return self.client

    @property
    def request_kwargs(self) -> Dict[str, Any]:
        """
        Returns the request parameters for the generative model.

        Returns:
            Dict[str, Any]: Request parameters for the generative model.
        """
        _request_params: Dict[str, Any] = {}
        if self.generation_config:
            _request_params["generation_config"] = self.generation_config
        if self.safety_settings:
            _request_params["safety_settings"] = self.safety_settings
        if self.generative_model_request_params:
            _request_params.update(self.generative_model_request_params)
        if self.function_declarations:
            _request_params["tools"] = [GeminiTool(function_declarations=self.function_declarations)]
        return _request_params

    def format_messages(self, messages: List[Message]) -> List[Content]:
        """
        Converts a list of Message objects to Gemini-compatible Content objects.

        Args:
            messages: List of Message objects containing various types of content

        Returns:
            List of Content objects formatted for Gemini's API
        """
        formatted_messages: List[Content] = []

        for msg in messages:
            if hasattr(msg, "response_tool_call_block"):
                formatted_messages.append(Content(role=msg.role, parts=msg.response_tool_call_block.parts))
                continue
            if msg.role == "tool" and hasattr(msg, "tool_call_result"):
                formatted_messages.append(msg.tool_call_result)
                continue
            if isinstance(msg.content, str):
                parts = [Part.from_text(msg.content)]
            elif isinstance(msg.content, list):
                parts = [Part.from_text(part) for part in msg.content if isinstance(part, str)]
            else:
                parts = []
            role = "model" if msg.role == "system" else "user" if msg.role == "tool" else msg.role

            formatted_messages.append(Content(role=role, parts=parts))

        return formatted_messages

    def format_functions(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """
        Converts function parameters to a Gemini-compatible format.

        Args:
            params (Dict[str, Any]): The original parameter's dictionary.

        Returns:
            Dict[str, Any]: The converted parameters dictionary compatible with Gemini.
        """
        formatted_params = {}
        for key, value in params.items():
            if key == "properties" and isinstance(value, dict):
                converted_properties = {}
                for prop_key, prop_value in value.items():
                    property_type = prop_value.get("type")
                    if isinstance(property_type, list):
                        # Create a copy to avoid modifying the original list
                        non_null_types = [t for t in property_type if t != "null"]
                        if non_null_types:
                            # Use the first non-null type
                            converted_type = non_null_types[0]
                        else:
                            # Default type if all types are 'null'
                            converted_type = "string"
                    else:
                        converted_type = property_type

                    converted_properties[prop_key] = {"type": converted_type}
                formatted_params[key] = converted_properties
            else:
                formatted_params[key] = value
        return formatted_params

    def add_tool(
            self,
            tool: Union["ModelTool", "Tool", Callable, dict, "Function"],
            strict: bool = False,
            agent: Optional[Any] = None,
    ) -> None:
        """
        Adds tools to the model.

        Args:
            tool: The tool to add. Can be a Tool, Toolkit, Callable, dict, or Function.
        """
        if self.function_declarations is None:
            self.function_declarations = []

        # If the tool is a Tool or Dict, log a warning.
        if isinstance(tool, ModelTool) or isinstance(tool, Dict):
            logger.warning("Tool of type 'Tool' or 'dict' is not yet supported by Gemini.")

        # If the tool is a Callable or Toolkit, add its functions to the Model
        elif callable(tool) or isinstance(tool, Tool) or isinstance(tool, Function):
            if self.functions is None:
                self.functions = {}

            if isinstance(tool, Tool):
                # For each function in the toolkit, process entrypoint and add to self.tools
                for name, func in tool.functions.items():
                    # If the function does not exist in self.functions, add to self.tools
                    if name not in self.functions:
                        func._agent = agent
                        func.process_entrypoint()
                        self.functions[name] = func
                        function_declaration = FunctionDeclaration(
                            name=func.name,
                            description=func.description,
                            parameters=self.format_functions(func.parameters),
                        )
                        self.function_declarations.append(function_declaration)
                        logger.debug(f"Function {name} from {tool.name} added to model.")

            elif isinstance(tool, Function):
                if tool.name not in self.functions:
                    tool._agent = agent
                    tool.process_entrypoint()
                    self.functions[tool.name] = tool
                    function_declaration = FunctionDeclaration(
                        name=tool.name,
                        description=tool.description,
                        parameters=self.format_functions(tool.parameters),
                    )
                    self.function_declarations.append(function_declaration)
                    logger.debug(f"Function {tool.name} added to model.")

            elif callable(tool):
                try:
                    function_name = tool.__name__
                    if function_name not in self.functions:
                        func = Function.from_callable(tool)
                        self.functions[func.name] = func
                        function_declaration = FunctionDeclaration(
                            name=func.name,
                            description=func.description,
                            parameters=self.format_functions(func.parameters),
                        )
                        self.function_declarations.append(function_declaration)
                        logger.debug(f"Function '{func.name}' added to model.")
                except Exception as e:
                    logger.warning(f"Could not add function {tool}: {e}")

    def invoke(self, messages: List[Message]) -> GenerationResponse:
        """
        Send a generate content request to VertexAI and return the response.

        Args:
            messages: List of Message objects containing various types of content

        Returns:
            GenerationResponse object containing the response content
        """
        return self.get_client().generate_content(contents=self.format_messages(messages))

    def invoke_stream(self, messages: List[Message]) -> Iterator[GenerationResponse]:
        """
        Send a generate content request to VertexAI and return the response.

        Args:
            messages: List of Message objects containing various types of content

        Returns:
            Iterator[GenerationResponse] object containing the response content
        """
        yield from self.get_client().generate_content(
            contents=self.format_messages(messages),
            stream=True,
        )

    def update_usage_metrics(
            self,
            assistant_message: Message,
            metrics: Metrics,
            usage: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Update usage metrics for the assistant message.

        Args:
            assistant_message: Message object containing the response content
            metrics: Metrics object containing the usage metrics
            usage: Dict[str, Any] object containing the usage metrics
        """
        assistant_message.metrics["time"] = metrics.response_timer.elapsed
        self.metrics.setdefault("response_times", []).append(metrics.response_timer.elapsed)
        if usage:
            metrics.input_tokens = usage.prompt_token_count or 0  # type: ignore
            metrics.output_tokens = usage.candidates_token_count or 0  # type: ignore
            metrics.total_tokens = usage.total_token_count or 0  # type: ignore

            if metrics.input_tokens is not None:
                assistant_message.metrics["input_tokens"] = metrics.input_tokens
                self.metrics["input_tokens"] = self.metrics.get("input_tokens", 0) + metrics.input_tokens
            if metrics.output_tokens is not None:
                assistant_message.metrics["output_tokens"] = metrics.output_tokens
                self.metrics["output_tokens"] = self.metrics.get("output_tokens", 0) + metrics.output_tokens
            if metrics.total_tokens is not None:
                assistant_message.metrics["total_tokens"] = metrics.total_tokens
                self.metrics["total_tokens"] = self.metrics.get("total_tokens", 0) + metrics.total_tokens
            if metrics.time_to_first_token is not None:
                assistant_message.metrics["time_to_first_token"] = metrics.time_to_first_token
                self.metrics.setdefault("time_to_first_token", []).append(metrics.time_to_first_token)

    def create_assistant_message(self, response: GenerationResponse, metrics: Metrics) -> Message:
        """
        Create an assistant message from the GenerationResponse.

        Args:
            response: GenerationResponse object containing the response content
            metrics: Metrics object containing the usage metrics

        Returns:
            Message object containing the assistant message
        """
        message_data = MessageData()

        message_data.response_candidates = response.candidates
        message_data.response_block = response.candidates[0].content
        message_data.response_role = message_data.response_block.role
        message_data.response_parts = message_data.response_block.parts
        message_data.response_usage = response.usage_metadata

        # -*- Parse response
        if message_data.response_parts is not None:
            for part in message_data.response_parts:
                part_dict = type(part).to_dict(part)

                # Extract text if present
                if "text" in part_dict:
                    message_data.response_content = part_dict.get("text")

                # Parse function calls
                if "function_call" in part_dict:
                    message_data.response_tool_call_block = response.candidates[0].content
                    message_data.response_tool_calls.append(
                        {
                            "type": "function",
                            "function": {
                                "name": part_dict.get("function_call").get("name"),
                                "arguments": json.dumps(part_dict.get("function_call").get("args")),
                            },
                        }
                    )

        # -*- Create assistant message
        assistant_message = Message(
            role=message_data.response_role or "model",
            content=message_data.response_content,
            response_tool_call_block=message_data.response_tool_call_block,
        )

        # -*- Update assistant message if tool calls are present
        if len(message_data.response_tool_calls) > 0:
            assistant_message.tool_calls = message_data.response_tool_calls

        # -*- Update usage metrics
        self.update_usage_metrics(
            assistant_message=assistant_message, metrics=metrics, usage=message_data.response_usage
        )

        return assistant_message

    def get_function_calls_to_run(
            self,
            assistant_message: Message,
            messages: List[Message],
    ) -> List[FunctionCall]:
        """
        Extracts and validates function calls from the assistant message.

        Args:
            assistant_message (Message): The assistant message containing tool calls.
            messages (List[Message]): The list of conversation messages.

        Returns:
            List[FunctionCall]: A list of valid function calls to run.
        """
        function_calls_to_run: List[FunctionCall] = []
        if assistant_message.tool_calls:
            for tool_call in assistant_message.tool_calls:
                _function_call = get_function_call_for_tool_call(tool_call, self.functions)
                if _function_call is None:
                    messages.append(Message(role="tool", content="Could not find function to call."))
                    continue
                if _function_call.error is not None:
                    messages.append(Message(role="tool", content=_function_call.error))
                    continue
                function_calls_to_run.append(_function_call)
        return function_calls_to_run

    def format_function_call_results(
            self,
            function_call_results: List[Message],
            messages: List[Message],
    ):
        """
        Processes the results of function calls and appends them to messages.

        Args:
            function_call_results (List[Message]): The results from running function calls.
            messages (List[Message]): The list of conversation messages.
        """
        if function_call_results:
            contents, parts = zip(
                *[
                    (
                        result.content,
                        Part.from_function_response(name=result.tool_name, response={"content": result.content}),
                    )
                    for result in function_call_results
                ]
            )

            messages.append(Message(role="tool", content=list(contents), tool_call_result=Content(parts=list(parts))))

    def handle_tool_calls(self, assistant_message: Message, messages: List[Message], model_response: ModelResponse):
        """
        Handle tool calls in the assistant message.

        Args:
            assistant_message (Message): The assistant message.
            messages (List[Message]): A list of messages.
            model_response (ModelResponse): The model response.

        Returns:
            Optional[ModelResponse]: The updated model response.
        """
        if assistant_message.tool_calls and self.run_tools:
            model_response.content = assistant_message.get_content_string() or ""
            function_calls_to_run = self.get_function_calls_to_run(assistant_message, messages)

            if self.show_tool_calls:
                if len(function_calls_to_run) == 1:
                    model_response.content += f"\n - Running: {function_calls_to_run[0].get_call_str()}\n\n"
                elif len(function_calls_to_run) > 1:
                    model_response.content += "\nRunning:"
                    for _f in function_calls_to_run:
                        model_response.content += f"\n - {_f.get_call_str()}"
                    model_response.content += "\n\n"

            function_call_results: List[Message] = []
            for _ in self.run_function_calls(
                    function_calls=function_calls_to_run,
                    function_call_results=function_call_results,
            ):
                pass

            self.format_function_call_results(function_call_results, messages)

            return model_response
        return None

    def response(self, messages: List[Message]) -> ModelResponse:
        """
        Send a generate content request to VertexAI and return the response.

        Args:
            messages: List of Message objects containing various types of content

        Returns:
            ModelResponse object containing the response content
        """
        self._log_messages(messages)
        model_response = ModelResponse()
        metrics = Metrics()

        metrics.response_timer.start()
        response: GenerationResponse = self.invoke(messages=messages)
        metrics.response_timer.stop()

        # -*- Create assistant message
        assistant_message = self.create_assistant_message(response=response, metrics=metrics)
        messages.append(assistant_message)

        # -*- Log response and metrics
        assistant_message.log()
        metrics.log()

        # -*- Handle tool calls
        if self.handle_tool_calls(assistant_message, messages, model_response):
            response_after_tool_calls = self.response(messages=messages)
            if response_after_tool_calls.content is not None:
                if model_response.content is None:
                    model_response.content = ""
                model_response.content += response_after_tool_calls.content
            return model_response

        # -*- Update model response
        if assistant_message.content is not None:
            model_response.content = assistant_message.get_content_string()
        if assistant_message.reasoning_content is not None:
            # add the reasoning content to the model response
            model_response.reasoning_content = assistant_message.reasoning_content

        # -*- Remove tool call blocks and tool call results from messages
        for m in messages:
            if hasattr(m, "response_tool_call_block"):
                m.response_tool_call_block = None
            if hasattr(m, "tool_call_result"):
                m.tool_call_result = None

        return model_response

    def handle_stream_tool_calls(self, assistant_message: Message, messages: List[Message]):
        """
        Parse and run function calls and append the results to messages.

        Args:
            assistant_message (Message): The assistant message containing tool calls.
            messages (List[Message]): The list of conversation messages.

        Yields:
            Iterator[ModelResponse]: Yields model responses during function execution.
        """
        if assistant_message.tool_calls and self.run_tools:
            function_calls_to_run = self.get_function_calls_to_run(assistant_message, messages)

            if self.show_tool_calls:
                if len(function_calls_to_run) == 1:
                    yield ModelResponse(content=f"\n - Running: {function_calls_to_run[0].get_call_str()}\n\n")
                elif len(function_calls_to_run) > 1:
                    yield ModelResponse(content="\nRunning:")
                    for _f in function_calls_to_run:
                        yield ModelResponse(content=f"\n - {_f.get_call_str()}")
                    yield ModelResponse(content="\n\n")

            function_call_results: List[Message] = []
            for intermediate_model_response in self.run_function_calls(
                    function_calls=function_calls_to_run, function_call_results=function_call_results
            ):
                yield intermediate_model_response

            self.format_function_call_results(function_call_results, messages)

    def response_stream(self, messages: List[Message]) -> Iterator[ModelResponse]:
        """
        Send a generate content request to VertexAI and return the response.

        Args:
            messages: List of Message objects containing various types of content

        Yields:
            Iterator[ModelResponse]: Yields model responses during function execution
        """
        self._log_messages(messages)
        message_data = MessageData()
        metrics = Metrics()

        metrics.response_timer.start()
        for response in self.invoke_stream(messages=messages):
            # -*- Parse response
            message_data.response_block = response.candidates[0].content
            if message_data.response_block is not None:
                metrics.time_to_first_token = metrics.response_timer.elapsed
            message_data.response_role = message_data.response_block.role
            if message_data.response_block.parts:
                message_data.response_parts = message_data.response_block.parts

            if message_data.response_parts is not None:
                for part in message_data.response_parts:
                    part_dict = type(part).to_dict(part)

                    # -*- Yield text if present
                    if "text" in part_dict:
                        text = part_dict.get("text")
                        yield ModelResponse(content=text)
                        message_data.response_content += text

                    # -*- Skip function calls if there are no parts
                    if not message_data.response_block.parts and message_data.response_parts:
                        continue
                    # -*- Parse function calls
                    if "function_call" in part_dict:
                        message_data.response_tool_call_block = response.candidates[0].content
                        message_data.response_tool_calls.append(
                            {
                                "type": "function",
                                "function": {
                                    "name": part_dict.get("function_call").get("name"),
                                    "arguments": json.dumps(part_dict.get("function_call").get("args")),
                                },
                            }
                        )
            message_data.response_usage = response.usage_metadata

        metrics.response_timer.stop()

        # -*- Create assistant message
        assistant_message = Message(
            role=message_data.response_role or "assistant",
            content=message_data.response_content,
            response_tool_call_block=message_data.response_tool_call_block,
        )

        # -*-  Update assistant message if tool calls are present
        if len(message_data.response_tool_calls) > 0:
            assistant_message.tool_calls = message_data.response_tool_calls

        self.update_usage_metrics(assistant_message=assistant_message, metrics=metrics,
                                  usage=message_data.response_usage)  # noqa

        # -*- Add assistant message to messages
        messages.append(assistant_message)

        # -*- Log response and metrics
        assistant_message.log()
        metrics.log()

        if assistant_message.tool_calls is not None and len(assistant_message.tool_calls) > 0 and self.run_tools:
            yield from self.handle_stream_tool_calls(assistant_message, messages)
            yield from self.response_stream(messages=messages)

        # -*- Remove tool call blocks and tool call results from messages
        for m in messages:
            if hasattr(m, "response_tool_call_block"):
                m.response_tool_call_block = None
            if hasattr(m, "tool_call_result"):
                m.tool_call_result = None



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
part of the code from https://github.com/phidatahq/phidata
"""
from os import getenv
from dataclasses import dataclass
from typing import Optional, Dict, List, Tuple, Any, Literal

from openai import OpenAI as OpenAIClient
from openai.types.create_embedding_response import CreateEmbeddingResponse

from agentica.emb.base import Emb
from agentica.utils.log import logger


@dataclass
class OpenAIEmb(Emb):
    model: str = "text-embedding-3-small"
    dimensions: int = 1536
    encoding_format: Literal["float", "base64"] = "float"
    user: Optional[str] = None
    api_key: Optional[str] = getenv("OPENAI_API_KEY")
    organization: Optional[str] = None
    base_url: Optional[str] = getenv("OPENAI_BASE_URL")
    request_params: Optional[Dict[str, Any]] = None
    client_params: Optional[Dict[str, Any]] = None
    openai_client: Optional[OpenAIClient] = None


    @property
    def client(self) -> OpenAIClient:
        if self.openai_client:
            return self.openai_client

        _client_params: Dict[str, Any] = {}
        if self.api_key:
            _client_params["api_key"] = self.api_key
        if self.organization:
            _client_params["organization"] = self.organization
        if self.base_url:
            _client_params["base_url"] = self.base_url
        if self.client_params:
            _client_params.update(self.client_params)
        self.openai_client = OpenAIClient(**_client_params)
        return self.openai_client

    def _response(self, text: str) -> CreateEmbeddingResponse:
        _request_params: Dict[str, Any] = {
            "input": text,
            "model": self.model,
            "encoding_format": self.encoding_format,
        }
        if self.user is not None:
            _request_params["user"] = self.user
        if self.model.startswith("text-embedding-3"):
            _request_params["dimensions"] = self.dimensions
        if self.request_params:
            _request_params.update(self.request_params)
        return self.client.embeddings.create(**_request_params)

    def get_embedding(self, text: str) -> List[float]:
        response: CreateEmbeddingResponse = self._response(text=text)
        try:
            return response.data[0].embedding
        except Exception as e:
            logger.warning(e)
            return []

    def get_embedding_and_usage(self, text: str) -> Tuple[List[float], Optional[Dict]]:
        response: CreateEmbeddingResponse = self._response(text=text)

        embedding = response.data[0].embedding
        usage = response.usage
        if usage:
            return embedding, usage.model_dump()
        return embedding, None

    def get_embeddings(self, texts: List[str]) -> List[List[float]]:
        return [self.get_embedding(text) for text in texts]


if __name__ == '__main__':
    emb = OpenAIEmb()
    text = "I love you"
    r = emb.get_embedding(text)
    print(r)
    r = emb.get_embedding_and_usage(text)
    print(r)
    texts = ["I love you", "我喜欢你"]
    r = emb.get_embeddings(texts)
    print(r)



try:
    from .server import (
        MCPServer,
        MCPServerSse,
        MCPServerSseParams,
        MCPServerStdio,
        MCPServerStdioParams,
        MCPServerStreamableHttp,
        MCPServerStreamableHttpParams,
    )
    from .client import MCPClient
except ImportError:
    pass

__all__ = [
    "MCPServer",
    "MCPServerSse",
    "MCPServerSseParams",
    "MCPServerStdio",
    "MCPServerStdioParams",
    "MCPServerStreamableHttp",
    "MCPServerStreamableHttpParams",
    "MCPClient",
]



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
part of the code from https://github.com/phidatahq/phidata
"""
from __future__ import annotations
import collections.abc
import inspect

from os import getenv
from uuid import uuid4
from types import GeneratorType
from typing import Any, Optional, Dict, Callable
from pydantic import BaseModel, Field, ConfigDict, field_validator, PrivateAttr

from agentica.utils.log import logger, set_log_level_to_debug
from agentica.agent import Agent
from agentica.run_response import RunResponse
from agentica.memory import WorkflowMemory, WorkflowRun
from agentica.storage.workflow.base import WorkflowStorage
from agentica.utils.misc import merge_dictionaries
from agentica.workflow_session import WorkflowSession


class Workflow(BaseModel):
    # -*- Workflow settings
    # Workflow name
    name: Optional[str] = None
    # Workflow description
    description: Optional[str] = None
    # Workflow UUID (autogenerated if not set)
    workflow_id: Optional[str] = Field(None, validate_default=True)
    # Metadata associated with this workflow
    workflow_data: Optional[Dict[str, Any]] = None

    # -*- User settings
    # ID of the user interacting with this workflow
    user_id: Optional[str] = None
    # Metadata associated with the user interacting with this workflow
    user_data: Optional[Dict[str, Any]] = None

    # -*- Session settings
    # Session UUID (autogenerated if not set)
    session_id: Optional[str] = Field(None, validate_default=True)
    # Session name
    session_name: Optional[str] = None
    # Session state stored in the database
    session_state: Dict[str, Any] = Field(default_factory=dict)

    # -*- Workflow Memory
    memory: WorkflowMemory = WorkflowMemory()

    # -*- Workflow Storage
    storage: Optional[WorkflowStorage] = None
    # WorkflowSession from the database: DO NOT SET MANUALLY
    _workflow_session: Optional[WorkflowSession] = None

    # debug_mode=True enables debug logs
    debug_mode: bool = Field(False, validate_default=True)
    # monitoring=True logs workflow information to phidata.com
    monitoring: bool = getenv("PHI_MONITORING", "false").lower() == "true"
    # telemetry=True logs minimal telemetry for analytics
    # This helps us improve the Agent and provide better support
    telemetry: bool = getenv("PHI_TELEMETRY", "true").lower() == "true"

    # DO NOT SET THE FOLLOWING FIELDS MANUALLY
    # Run ID: DO NOT SET MANUALLY
    run_id: Optional[str] = None
    # Input to the Workflow run: DO NOT SET MANUALLY
    run_input: Optional[Dict[str, Any]] = None
    # Response from the Workflow run: DO NOT SET MANUALLY
    run_response: RunResponse = Field(default_factory=RunResponse)
    # Metadata associated with this session: DO NOT SET MANUALLY
    session_data: Optional[Dict[str, Any]] = None

    # The run function provided by the subclass
    _subclass_run: Callable = PrivateAttr()
    # Parameters of the run function
    _run_parameters: Dict[str, Any] = PrivateAttr()
    # Return type of the run function
    _run_return_type: Optional[str] = PrivateAttr()

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    @field_validator("workflow_id", mode="before")
    def set_workflow_id(cls, v: Optional[str]) -> str:
        workflow_id = v or str(uuid4())
        logger.debug(f"*********** Worfklow ID: {workflow_id} ***********")
        return workflow_id

    @field_validator("session_id", mode="before")
    def set_session_id(cls, v: Optional[str]) -> str:
        session_id = v or str(uuid4())
        logger.debug(f"*********** Worflow Session ID: {session_id} ***********")
        return session_id

    @field_validator("debug_mode", mode="before")
    def set_log_level(cls, v: bool) -> bool:
        if v or getenv("PHI_DEBUG", "false").lower() == "true":
            set_log_level_to_debug()
            logger.debug("Debug logs enabled")
        return v

    def get_workflow_data(self) -> Dict[str, Any]:
        workflow_data = self.workflow_data or {}
        if self.name is not None:
            workflow_data["name"] = self.name
        return workflow_data

    def get_session_data(self) -> Dict[str, Any]:
        session_data = self.session_data or {}
        if self.session_name is not None:
            session_data["session_name"] = self.session_name
        if len(self.session_state) > 0:
            session_data["session_state"] = self.session_state
        return session_data

    def get_workflow_session(self) -> WorkflowSession:
        """Get a WorkflowSession object, which can be saved to the database"""

        return WorkflowSession(
            session_id=self.session_id,
            workflow_id=self.workflow_id,
            user_id=self.user_id,
            memory=self.memory.to_dict(),
            workflow_data=self.get_workflow_data(),
            user_data=self.user_data,
            session_data=self.get_session_data(),
        )

    def from_workflow_session(self, session: WorkflowSession):
        """Load the existing Workflow from a WorkflowSession (from the database)"""

        # Get the session_id, workflow_id and user_id from the database
        if self.session_id is None and session.session_id is not None:
            self.session_id = session.session_id
        if self.workflow_id is None and session.workflow_id is not None:
            self.workflow_id = session.workflow_id
        if self.user_id is None and session.user_id is not None:
            self.user_id = session.user_id

        # Read workflow_data from the database
        if session.workflow_data is not None:
            # Get name from database and update the workflow name if not set
            if self.name is None and "name" in session.workflow_data:
                self.name = session.workflow_data.get("name")

            # If workflow_data is set in the workflow, update the database workflow_data with the workflow's workflow_data
            if self.workflow_data is not None:
                # Updates workflow_session.workflow_data in place
                merge_dictionaries(session.workflow_data, self.workflow_data)
            self.workflow_data = session.workflow_data

        # Read user_data from the database
        if session.user_data is not None:
            # If user_data is set in the workflow, update the database user_data with the workflow's user_data
            if self.user_data is not None:
                # Updates workflow_session.user_data in place
                merge_dictionaries(session.user_data, self.user_data)
            self.user_data = session.user_data

        # Read session_data from the database
        if session.session_data is not None:
            # Get the session_name from database and update the current session_name if not set
            if self.session_name is None and "session_name" in session.session_data:
                self.session_name = session.session_data.get("session_name")

            # Get the session_state from database and update the current session_state
            if "session_state" in session.session_data:
                session_state_from_db = session.session_data.get("session_state")
                if (
                        session_state_from_db is not None
                        and isinstance(session_state_from_db, dict)
                        and len(session_state_from_db) > 0
                ):
                    # If the session_state is already set, merge the session_state from the database with the current session_state
                    if len(self.session_state) > 0:
                        # This updates session_state_from_db
                        merge_dictionaries(session_state_from_db, self.session_state)
                    # Update the current session_state
                    self.session_state = session_state_from_db

            # If session_data is set in the workflow, update the database session_data with the workflow's session_data
            if self.session_data is not None:
                # Updates workflow_session.session_data in place
                merge_dictionaries(session.session_data, self.session_data)
            self.session_data = session.session_data

        # Read memory from the database
        if session.memory is not None:
            try:
                if "runs" in session.memory:
                    self.memory.runs = [WorkflowRun(**m) for m in session.memory["runs"]]
            except Exception as e:
                logger.warning(f"Failed to load WorkflowMemory: {e}")
        logger.debug(f"-*- WorkflowSession loaded: {session.session_id}")

    def read_from_storage(self) -> Optional[WorkflowSession]:
        """Load the WorkflowSession from storage.

        Returns:
            Optional[WorkflowSession]: The loaded WorkflowSession or None if not found.
        """
        if self.storage is not None and self.session_id is not None:
            self._workflow_session = self.storage.read(session_id=self.session_id)
            if self._workflow_session is not None:
                self.from_workflow_session(session=self._workflow_session)
        return self._workflow_session

    def write_to_storage(self) -> Optional[WorkflowSession]:
        """Save the WorkflowSession to storage

        Returns:
            Optional[WorkflowSession]: The saved WorkflowSession or None if not saved.
        """
        if self.storage is not None:
            self._workflow_session = self.storage.upsert(session=self.get_workflow_session())
        return self._workflow_session

    def load_session(self, force: bool = False) -> Optional[str]:
        """Load an existing session from the database and return the session_id.
        If a session does not exist, create a new session.

        - If a session exists in the database, load the session.
        - If a session does not exist in the database, create a new session.
        """
        # If a workflow_session is already loaded, return the session_id from the workflow_session
        # if session_id matches the session_id from the workflow_session
        if self._workflow_session is not None and not force:
            if self.session_id is not None and self._workflow_session.session_id == self.session_id:
                return self._workflow_session.session_id

        # Load an existing session or create a new session
        if self.storage is not None:
            # Load existing session if session_id is provided
            logger.debug(f"Reading WorkflowSession: {self.session_id}")
            self.read_from_storage()

            # Create a new session if it does not exist
            if self._workflow_session is None:
                logger.debug("-*- Creating new WorkflowSession")
                # write_to_storage() will create a new WorkflowSession
                # and populate self._workflow_session with the new session
                self.write_to_storage()
                if self._workflow_session is None:
                    raise Exception("Failed to create new WorkflowSession in storage")
                logger.debug(f"-*- Created WorkflowSession: {self._workflow_session.session_id}")
                self.log_workflow_session()
        return self.session_id

    def run(self, *args: Any, **kwargs: Any):
        logger.error(f"{self.__class__.__name__}.run() method not implemented.")
        return

    def run_workflow(self, *args: Any, **kwargs: Any):
        self.run_id = str(uuid4())
        self.run_input = {"args": args, "kwargs": kwargs}
        self.run_response = RunResponse(run_id=self.run_id, session_id=self.session_id, workflow_id=self.workflow_id)
        self.read_from_storage()
        result = self._subclass_run(*args, **kwargs)

        # Case 1: The run method returns an Iterator[RunResponse]
        if isinstance(result, (GeneratorType, collections.abc.Iterator)):
            # Initialize the run_response content
            self.run_response.content = ""

            def result_generator():
                for item in result:
                    if isinstance(item, RunResponse):
                        # Update the run_id, session_id and workflow_id of the RunResponse
                        item.run_id = self.run_id
                        item.session_id = self.session_id
                        item.workflow_id = self.workflow_id

                        # Update the run_response with the content from the result
                        if item.content is not None and isinstance(item.content, str):
                            self.run_response.content += item.content
                    else:
                        logger.warning(f"Workflow.run() should only yield RunResponse objects, got: {type(item)}")
                    yield item

                # Add the run to the memory
                self.memory.add_run(WorkflowRun(input=self.run_input, response=self.run_response))
                # Write this run to the database
                self.write_to_storage()
            return result_generator()
        # Case 2: The run method returns a RunResponse
        elif isinstance(result, RunResponse):
            # Update the result with the run_id, session_id and workflow_id of the workflow run
            result.run_id = self.run_id
            result.session_id = self.session_id
            result.workflow_id = self.workflow_id

            # Update the run_response with the content from the result
            if result.content is not None and isinstance(result.content, str):
                self.run_response.content = result.content

            # Add the run to the memory
            self.memory.add_run(WorkflowRun(input=self.run_input, response=self.run_response))
            # Write this run to the database
            self.write_to_storage()
            return result
        else:
            logger.warning(f"Workflow.run() should only return RunResponse objects, got: {type(result)}")
            return None

    def __init__(self, **data):
        super().__init__(**data)
        self.name = self.name or self.__class__.__name__
        # Check if 'run' is provided by the subclass
        if self.__class__.run is not Workflow.run:
            # Store the original run method bound to the instance
            self._subclass_run = self.__class__.run.__get__(self)
            # Get the parameters of the run method
            sig = inspect.signature(self.__class__.run)
            # Convert parameters to a serializable format
            self._run_parameters = {
                name: {
                    "name": name,
                    "default": param.default if param.default is not inspect.Parameter.empty else None,
                    "annotation": (
                        param.annotation.__name__
                        if hasattr(param.annotation, "__name__")
                        else (
                            str(param.annotation).replace("typing.Optional[", "").replace("]", "")
                            if "typing.Optional" in str(param.annotation)
                            else str(param.annotation)
                        )
                    )
                    if param.annotation is not inspect.Parameter.empty
                    else None,
                    "required": param.default is inspect.Parameter.empty,
                }
                for name, param in sig.parameters.items()
                if name != "self"
            }
            # Determine the return type of the run method
            return_annotation = sig.return_annotation
            self._run_return_type = (
                return_annotation.__name__
                if return_annotation is not inspect.Signature.empty and hasattr(return_annotation, "__name__")
                else str(return_annotation)
                if return_annotation is not inspect.Signature.empty
                else None
            )
            # Replace the instance's run method with run_workflow
            object.__setattr__(self, "run", self.run_workflow.__get__(self))
        else:
            # This will log an error when called
            self._subclass_run = self.run
            self._run_parameters = {}
            self._run_return_type = None

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)
        for field_name, field in self.__fields__.items():
            value = getattr(self, field_name)
            if isinstance(value, Agent):
                value.session_id = self.session_id

    def log_workflow_session(self):
        logger.debug(f"*********** Logging WorkflowSession: {self.session_id} ***********")

    def rename_session(self, session_id: str, name: str):
        if self.storage is None:
            raise ValueError("Storage is not set")
        workflow_session = self.storage.read(session_id)
        if workflow_session is None:
            raise Exception(f"WorkflowSession not found: {session_id}")
        if workflow_session.session_data is not None:
            workflow_session.session_data["session_name"] = name
        else:
            workflow_session.session_data = {"session_name": name}
        self.storage.upsert(workflow_session)

    def delete_session(self, session_id: str):
        if self.storage is None:
            raise ValueError("Storage is not set")
        self.storage.delete_session(session_id)

    def deep_copy(self, *, update: Optional[Dict[str, Any]] = None) -> "Workflow":
        """Create and return a deep copy of this Workflow, optionally updating fields.

        Args:
            update (Optional[Dict[str, Any]]): Optional dictionary of fields for the new Workflow.

        Returns:
            Workflow: A new Workflow instance.
        """
        # Extract the fields to set for the new Workflow
        fields_for_new_workflow = {}

        for field_name in self.model_fields_set:
            field_value = getattr(self, field_name)
            if field_value is not None:
                if isinstance(field_value, Agent):
                    fields_for_new_workflow[field_name] = field_value.deep_copy()
                else:
                    fields_for_new_workflow[field_name] = self._deep_copy_field(field_name, field_value)

        # Update fields if provided
        if update:
            fields_for_new_workflow.update(update)

        # Create a new Workflow
        new_workflow = self.__class__(**fields_for_new_workflow)
        logger.debug(
            f"Created new Workflow: workflow_id: {new_workflow.workflow_id} | session_id: {new_workflow.session_id}"
        )
        return new_workflow

    def _deep_copy_field(self, field_name: str, field_value: Any) -> Any:
        """Helper method to deep copy a field based on its type."""
        from copy import copy, deepcopy

        # For memory, use its deep_copy method
        if field_name == "memory":
            return field_value.deep_copy()

        # For compound types, attempt a deep copy
        if isinstance(field_value, (list, dict, set, WorkflowStorage)):
            try:
                return deepcopy(field_value)
            except Exception as e:
                logger.warning(f"Failed to deepcopy field: {field_name} - {e}")
                try:
                    return copy(field_value)
                except Exception as e:
                    logger.warning(f"Failed to copy field: {field_name} - {e}")
                    return field_value

        # For pydantic models, attempt a deep copy
        if isinstance(field_value, BaseModel):
            try:
                return field_value.model_copy(deep=True)
            except Exception as e:
                logger.warning(f"Failed to deepcopy field: {field_name} - {e}")
                try:
                    return field_value.model_copy(deep=False)
                except Exception as e:
                    logger.warning(f"Failed to copy field: {field_name} - {e}")
                    return field_value

        # For other types, return as is
        return field_value



from typing import Optional
from os import getenv

from agentica.model.openai.like import OpenAILike


class Moonshot(OpenAILike):
    """
    A model class for Moonshot Chat API.

    Attributes:
    - id: str: The unique identifier of the model. Default: "moonshot-v1-auto".
    - name: str: The name of the model. Default: "MoonshotChat".
    - provider: str: The provider of the model. Default: "MoonShot".
    - api_key: Optional[str]: The API key for the model.
    - base_url: str: The base URL for the model. Default: "https://api.moonshot.cn/v1".
    """

    id: str = "moonshot-v1-8k"
    name: str = "Moonshot"
    provider: str = "MoonShot"

    api_key: Optional[str] = getenv("MOONSHOT_API_KEY", None)
    base_url: str = "https://api.moonshot.cn/v1"



from os import getenv
from dataclasses import dataclass, field
from typing import Optional, List, Iterator, Dict, Any, Union

import httpx

from agentica.model.base import Model
from agentica.model.message import Message
from agentica.model.response import ModelResponse
from agentica.tools.base import FunctionCall, get_function_call_for_tool_call
from agentica.utils.log import logger
from agentica.utils.timer import Timer

try:
    from groq import Groq as GroqClient, AsyncGroq as AsyncGroqClient
    from groq.types.chat import ChatCompletion, ChatCompletionMessage
    from groq.types.chat.chat_completion_chunk import ChatCompletionChunk, ChoiceDeltaToolCall, ChoiceDelta
    from groq.types.completion_usage import CompletionUsage
except (ModuleNotFoundError, ImportError):
    raise ImportError("`groq` not installed. Please install using `pip install groq`")


@dataclass
class Metrics:
    input_tokens: int = 0
    output_tokens: int = 0
    total_tokens: int = 0
    prompt_tokens: int = 0
    completion_tokens: int = 0
    completion_time: Optional[float] = None
    prompt_time: Optional[float] = None
    queue_time: Optional[float] = None
    total_time: Optional[float] = None
    time_to_first_token: Optional[float] = None
    response_timer: Timer = field(default_factory=Timer)

    def log(self):
        if self.time_to_first_token is not None:
            logger.debug(f"* Time to first token:         {self.time_to_first_token:.4f}s")
        logger.debug(f"* Time to generate response:   {self.response_timer.elapsed:.4f}s")
        logger.debug(f"* Tokens per second:           {self.output_tokens / self.response_timer.elapsed:.4f} tokens/s")
        logger.debug(f"* Input tokens:                {self.input_tokens or self.prompt_tokens}")
        logger.debug(f"* Output tokens:               {self.output_tokens or self.completion_tokens}")
        logger.debug(f"* Total tokens:                {self.total_tokens}")
        if self.completion_time is not None:
            logger.debug(f"* Completion time:             {self.completion_time:.4f}s")
        if self.prompt_time is not None:
            logger.debug(f"* Prompt time:                 {self.prompt_time:.4f}s")
        if self.queue_time is not None:
            logger.debug(f"* Queue time:                  {self.queue_time:.4f}s")
        if self.total_time is not None:
            logger.debug(f"* Total time:                  {self.total_time:.4f}s")


@dataclass
class StreamData:
    response_content: str = ""
    response_tool_calls: Optional[List[ChoiceDeltaToolCall]] = None


class Groq(Model):
    """
    A class for interacting with Groq models.

    For more information, see: https://console.groq.com/docs/libraries
    """

    id: str = "llama3-groq-70b-8192-tool-use-preview"
    name: str = "Groq"
    provider: str = "Groq"

    # Request parameters
    frequency_penalty: Optional[float] = None
    logit_bias: Optional[Any] = None
    logprobs: Optional[bool] = None
    max_tokens: Optional[int] = None
    presence_penalty: Optional[float] = None
    response_format: Optional[Dict[str, Any]] = None
    seed: Optional[int] = None
    stop: Optional[Union[str, List[str]]] = None
    temperature: Optional[float] = None
    top_logprobs: Optional[int] = None
    top_p: Optional[float] = None
    user: Optional[str] = None
    extra_headers: Optional[Any] = None
    extra_query: Optional[Any] = None
    request_params: Optional[Dict[str, Any]] = None

    # Client parameters
    api_key: Optional[str] = None
    base_url: Optional[Union[str, httpx.URL]] = None
    timeout: Optional[int] = None
    max_retries: Optional[int] = None
    default_headers: Optional[Any] = None
    default_query: Optional[Any] = None
    http_client: Optional[httpx.Client] = None
    client_params: Optional[Dict[str, Any]] = None

    # Groq clients
    client: Optional[GroqClient] = None
    async_client: Optional[AsyncGroqClient] = None

    def get_client_params(self) -> Dict[str, Any]:
        self.api_key = self.api_key or getenv("GROQ_API_KEY")
        if not self.api_key:
            logger.error("GROQ_API_KEY not set. Please set the GROQ_API_KEY environment variable.")

        client_params: Dict[str, Any] = {}
        if self.api_key:
            client_params["api_key"] = self.api_key
        if self.base_url:
            client_params["base_url"] = self.base_url
        if self.timeout:
            client_params["timeout"] = self.timeout
        if self.max_retries:
            client_params["max_retries"] = self.max_retries
        if self.default_headers:
            client_params["default_headers"] = self.default_headers
        if self.default_query:
            client_params["default_query"] = self.default_query
        if self.client_params:
            client_params.update(self.client_params)
        return client_params

    def get_client(self) -> GroqClient:
        """
        Returns a Groq client.

        Returns:
            GroqClient: An instance of the Groq client.
        """
        if self.client:
            return self.client

        client_params: Dict[str, Any] = self.get_client_params()
        if self.http_client is not None:
            client_params["http_client"] = self.http_client
        return GroqClient(**client_params)

    def get_async_client(self) -> AsyncGroqClient:
        """
        Returns an asynchronous Groq client.

        Returns:
            AsyncGroqClient: An instance of the asynchronous Groq client.
        """
        if self.async_client:
            return self.async_client

        client_params: Dict[str, Any] = self.get_client_params()
        if self.http_client:
            client_params["http_client"] = self.http_client
        else:
            # Create a new async HTTP client with custom limits
            client_params["http_client"] = httpx.AsyncClient(
                limits=httpx.Limits(max_connections=1000, max_keepalive_connections=100)
            )
        return AsyncGroqClient(**client_params)

    @property
    def request_kwargs(self) -> Dict[str, Any]:
        """
        Returns keyword arguments for API requests.

        Returns:
            Dict[str, Any]: A dictionary of keyword arguments for API requests.
        """
        request_params: Dict[str, Any] = {}
        if self.frequency_penalty:
            request_params["frequency_penalty"] = self.frequency_penalty
        if self.logit_bias:
            request_params["logit_bias"] = self.logit_bias
        if self.logprobs:
            request_params["logprobs"] = self.logprobs
        if self.max_tokens:
            request_params["max_tokens"] = self.max_tokens
        if self.presence_penalty:
            request_params["presence_penalty"] = self.presence_penalty
        if self.response_format:
            request_params["response_format"] = self.response_format
        if self.seed:
            request_params["seed"] = self.seed
        if self.stop:
            request_params["stop"] = self.stop
        if self.temperature:
            request_params["temperature"] = self.temperature
        if self.top_logprobs:
            request_params["top_logprobs"] = self.top_logprobs
        if self.top_p:
            request_params["top_p"] = self.top_p
        if self.user:
            request_params["user"] = self.user
        if self.extra_headers:
            request_params["extra_headers"] = self.extra_headers
        if self.extra_query:
            request_params["extra_query"] = self.extra_query
        if self.tools:
            request_params["tools"] = self.get_tools_for_api()
            if self.tool_choice is None:
                request_params["tool_choice"] = "auto"
            else:
                request_params["tool_choice"] = self.tool_choice
        if self.request_params:
            request_params.update(self.request_params)
        return request_params

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the model to a dictionary.

        Returns:
            Dict[str, Any]: The dictionary representation of the model.
        """
        model_dict = super().to_dict()
        if self.frequency_penalty:
            model_dict["frequency_penalty"] = self.frequency_penalty
        if self.logit_bias:
            model_dict["logit_bias"] = self.logit_bias
        if self.logprobs:
            model_dict["logprobs"] = self.logprobs
        if self.max_tokens:
            model_dict["max_tokens"] = self.max_tokens
        if self.presence_penalty:
            model_dict["presence_penalty"] = self.presence_penalty
        if self.response_format:
            model_dict["response_format"] = self.response_format
        if self.seed:
            model_dict["seed"] = self.seed
        if self.stop:
            model_dict["stop"] = self.stop
        if self.temperature:
            model_dict["temperature"] = self.temperature
        if self.top_logprobs:
            model_dict["top_logprobs"] = self.top_logprobs
        if self.top_p:
            model_dict["top_p"] = self.top_p
        if self.user:
            model_dict["user"] = self.user
        if self.extra_headers:
            model_dict["extra_headers"] = self.extra_headers
        if self.extra_query:
            model_dict["extra_query"] = self.extra_query
        if self.tools:
            model_dict["tools"] = self.get_tools_for_api()
            if self.tool_choice is None:
                model_dict["tool_choice"] = "auto"
            else:
                model_dict["tool_choice"] = self.tool_choice
        return model_dict

    def format_message(self, message: Message) -> Dict[str, Any]:
        """
        Format a message into the format expected by OpenAI.

        Args:
            message (Message): The message to format.

        Returns:
            Dict[str, Any]: The formatted message.
        """
        if message.role == "user":
            if message.images is not None:
                message = self.add_images_to_message(message=message, images=message.images)
            if message.audio is not None:
                message = self.add_audio_to_message(message=message, audio=message.audio)

        return message.to_dict()

    def invoke(self, messages: List[Message]) -> ChatCompletion:
        """
        Send a chat completion request to the Groq API.

        Args:
            messages (List[Message]): A list of messages to send to the model.

        Returns:
            ChatCompletion: The chat completion response from the API.
        """
        return self.get_client().chat.completions.create(
            model=self.id,
            messages=[self.format_message(m) for m in messages],  # type: ignore
            **self.request_kwargs,
        )

    async def ainvoke(self, messages: List[Message]) -> ChatCompletion:
        """
        Sends an asynchronous chat completion request to the Groq API.

        Args:
            messages (List[Message]): A list of messages to send to the model.

        Returns:
            ChatCompletion: The chat completion response from the API.
        """
        return await self.get_async_client().chat.completions.create(
            model=self.id,
            messages=[self.format_message(m) for m in messages],  # type: ignore
            **self.request_kwargs,
        )

    def invoke_stream(self, messages: List[Message]) -> Iterator[ChatCompletionChunk]:
        """
        Send a streaming chat completion request to the Groq API.

        Args:
            messages (List[Message]): A list of messages to send to the model.

        Returns:
            Iterator[ChatCompletionChunk]: An iterator of chat completion chunks.
        """
        yield from self.get_client().chat.completions.create(
            model=self.id,
            messages=[self.format_message(m) for m in messages],  # type: ignore
            stream=True,
            **self.request_kwargs,
        )

    async def ainvoke_stream(self, messages: List[Message]) -> Any:
        """
        Sends an asynchronous streaming chat completion request to the Groq API.

        Args:
            messages (List[Message]): A list of messages to send to the model.

        Returns:
            Any: An asynchronous iterator of chat completion chunks.
        """
        async_stream = await self.get_async_client().chat.completions.create(
            model=self.id,
            messages=[self.format_message(m) for m in messages],  # type: ignore
            stream=True,
            **self.request_kwargs,
        )
        async for chunk in async_stream:  # type: ignore
            yield chunk

    def handle_tool_calls(
            self,
            assistant_message: Message,
            messages: List[Message],
            model_response: ModelResponse,
            tool_role: str = "tool",
    ) -> Optional[ModelResponse]:
        """
        Handle tool calls in the assistant message.

        Args:
            assistant_message (Message): The assistant message.
            messages (List[Message]): The list of messages.
            model_response (ModelResponse): The model response.
            tool_role (str): The role of the tool call. Defaults to "tool".

        Returns:
            Optional[ModelResponse]: The model response after handling tool calls.
        """
        if assistant_message.tool_calls is not None and len(assistant_message.tool_calls) > 0 and self.run_tools:
            if model_response.content is None:
                model_response.content = ""
            function_call_results: List[Message] = []
            function_calls_to_run: List[FunctionCall] = []
            for tool_call in assistant_message.tool_calls:
                _tool_call_id = tool_call.get("id")
                _function_call = get_function_call_for_tool_call(tool_call, self.functions)
                if _function_call is None:
                    messages.append(
                        Message(
                            role="tool",
                            tool_call_id=_tool_call_id,
                            content="Could not find function to call.",
                        )
                    )
                    continue
                if _function_call.error is not None:
                    messages.append(
                        Message(
                            role="tool",
                            tool_call_id=_tool_call_id,
                            content=_function_call.error,
                        )
                    )
                    continue
                function_calls_to_run.append(_function_call)

            if self.show_tool_calls:
                model_response.content += "\nRunning:"
                for _f in function_calls_to_run:
                    model_response.content += f"\n - {_f.get_call_str()}"
                model_response.content += "\n\n"

            for _ in self.run_function_calls(
                    function_calls=function_calls_to_run, function_call_results=function_call_results,
                    tool_role=tool_role
            ):
                pass

            if len(function_call_results) > 0:
                messages.extend(function_call_results)

            return model_response
        return None

    def update_usage_metrics(
            self, assistant_message: Message, metrics: Metrics, response_usage: Optional[CompletionUsage]
    ) -> None:
        """
        Update the usage metrics for the assistant message and the model.

        Args:
            assistant_message (Message): The assistant message.
            metrics (Metrics): The metrics.
            response_usage (Optional[CompletionUsage]): The response usage.
        """
        # Update time taken to generate response
        assistant_message.metrics["time"] = metrics.response_timer.elapsed
        self.metrics.setdefault("response_times", []).append(metrics.response_timer.elapsed)
        if response_usage:
            prompt_tokens = response_usage.prompt_tokens
            completion_tokens = response_usage.completion_tokens
            total_tokens = response_usage.total_tokens

            if prompt_tokens is not None:
                metrics.input_tokens = prompt_tokens
                metrics.prompt_tokens = prompt_tokens
                assistant_message.metrics["input_tokens"] = prompt_tokens
                assistant_message.metrics["prompt_tokens"] = prompt_tokens
                self.metrics["input_tokens"] = self.metrics.get("input_tokens", 0) + prompt_tokens
                self.metrics["prompt_tokens"] = self.metrics.get("prompt_tokens", 0) + prompt_tokens
            if completion_tokens is not None:
                metrics.output_tokens = completion_tokens
                metrics.completion_tokens = completion_tokens
                assistant_message.metrics["output_tokens"] = completion_tokens
                assistant_message.metrics["completion_tokens"] = completion_tokens
                self.metrics["output_tokens"] = self.metrics.get("output_tokens", 0) + completion_tokens
                self.metrics["completion_tokens"] = self.metrics.get("completion_tokens", 0) + completion_tokens
            if total_tokens is not None:
                metrics.total_tokens = total_tokens
                assistant_message.metrics["total_tokens"] = total_tokens
                self.metrics["total_tokens"] = self.metrics.get("total_tokens", 0) + total_tokens
            if response_usage.completion_time is not None:
                metrics.completion_time = response_usage.completion_time
                assistant_message.metrics["completion_time"] = response_usage.completion_time
                self.metrics["completion_time"] = (
                        self.metrics.get("completion_time", 0) + response_usage.completion_time
                )
            if response_usage.prompt_time is not None:
                metrics.prompt_time = response_usage.prompt_time
                assistant_message.metrics["prompt_time"] = response_usage.prompt_time
                self.metrics["prompt_time"] = self.metrics.get("prompt_time", 0) + response_usage.prompt_time
            if response_usage.queue_time is not None:
                metrics.queue_time = response_usage.queue_time
                assistant_message.metrics["queue_time"] = response_usage.queue_time
                self.metrics["queue_time"] = self.metrics.get("queue_time", 0) + response_usage.queue_time
            if response_usage.total_time is not None:
                metrics.total_time = response_usage.total_time
                assistant_message.metrics["total_time"] = response_usage.total_time
                self.metrics["total_time"] = self.metrics.get("total_time", 0) + response_usage.total_time

    def create_assistant_message(
            self,
            response_message: ChatCompletionMessage,
            metrics: Metrics,
            response_usage: Optional[CompletionUsage],
    ) -> Message:
        """
        Create an assistant message from the response.

        Args:
            response_message (ChatCompletionMessage): The response message.
            metrics (Metrics): The metrics.
            response_usage (Optional[CompletionUsage]): The response usage.

        Returns:
            Message: The assistant message.
        """
        assistant_message = Message(
            role=response_message.role or "assistant",
            content=response_message.content,
        )
        if response_message.tool_calls is not None and len(response_message.tool_calls) > 0:
            try:
                assistant_message.tool_calls = [t.model_dump() for t in response_message.tool_calls]
            except Exception as e:
                logger.warning(f"Error processing tool calls: {e}")

        # Update metrics
        self.update_usage_metrics(assistant_message, metrics, response_usage)
        return assistant_message

    def response(self, messages: List[Message]) -> ModelResponse:
        """
        Generate a response from Groq.

        Args:
            messages (List[Message]): A list of messages.

        Returns:
            ModelResponse: The model response.
        """
        self._log_messages(messages)
        model_response = ModelResponse()
        metrics = Metrics()

        # -*- Generate response
        metrics.response_timer.start()
        response: ChatCompletion = self.invoke(messages=messages)
        metrics.response_timer.stop()

        # -*- Parse response
        response_message: ChatCompletionMessage = response.choices[0].message
        response_usage: Optional[CompletionUsage] = response.usage

        # -*- Create assistant message
        assistant_message = self.create_assistant_message(
            response_message=response_message, metrics=metrics, response_usage=response_usage
        )

        # -*- Add assistant message to messages
        messages.append(assistant_message)

        # -*- Log response and metrics
        assistant_message.log()
        metrics.log()

        # -*- Update model response with assistant message content and audio
        if assistant_message.content is not None:
            # add the content to the model response
            model_response.content = assistant_message.get_content_string()

        # -*- Handle tool calls
        tool_role = "tool"
        if (
                self.handle_tool_calls(
                    assistant_message=assistant_message,
                    messages=messages,
                    model_response=model_response,
                    tool_role=tool_role,
                )
                is not None
        ):
            return self.handle_post_tool_call_messages(messages=messages, model_response=model_response)
        return model_response

    async def aresponse(self, messages: List[Message]) -> ModelResponse:
        """
        Generate an asynchronous response from Groq.

        Args:
            messages (List[Message]): A list of messages.

        Returns:
            ModelResponse: The model response from the API.
        """
        self._log_messages(messages)
        model_response = ModelResponse()
        metrics = Metrics()

        # -*- Generate response
        metrics.response_timer.start()
        response: ChatCompletion = await self.ainvoke(messages=messages)
        metrics.response_timer.stop()

        # -*- Parse response
        response_message: ChatCompletionMessage = response.choices[0].message
        response_usage: Optional[CompletionUsage] = response.usage

        # -*- Create assistant message
        assistant_message = self.create_assistant_message(
            response_message=response_message, metrics=metrics, response_usage=response_usage
        )

        # -*- Add assistant message to messages
        messages.append(assistant_message)

        # -*- Log response and metrics
        assistant_message.log()
        metrics.log()

        # -*- Update model response with assistant message content and audio
        if assistant_message.content is not None:
            # add the content to the model response
            model_response.content = assistant_message.get_content_string()

        # -*- Handle tool calls
        tool_role = "tool"
        if (
                self.handle_tool_calls(
                    assistant_message=assistant_message,
                    messages=messages,
                    model_response=model_response,
                    tool_role=tool_role,
                )
                is not None
        ):
            return await self.ahandle_post_tool_call_messages(messages=messages, model_response=model_response)

        return model_response

    def update_stream_metrics(self, assistant_message: Message, metrics: Metrics):
        """
        Update the usage metrics for the assistant message and the model.

        Args:
            assistant_message (Message): The assistant message.
            metrics (Metrics): The metrics.
        """
        # Update time taken to generate response
        assistant_message.metrics["time"] = metrics.response_timer.elapsed
        self.metrics.setdefault("response_times", []).append(metrics.response_timer.elapsed)

        if metrics.time_to_first_token is not None:
            assistant_message.metrics["time_to_first_token"] = metrics.time_to_first_token
            self.metrics.setdefault("time_to_first_token", []).append(metrics.time_to_first_token)

        if metrics.input_tokens is not None:
            assistant_message.metrics["input_tokens"] = metrics.input_tokens
            self.metrics["input_tokens"] = self.metrics.get("input_tokens", 0) + metrics.input_tokens
        if metrics.output_tokens is not None:
            assistant_message.metrics["output_tokens"] = metrics.output_tokens
            self.metrics["output_tokens"] = self.metrics.get("output_tokens", 0) + metrics.output_tokens
        if metrics.prompt_tokens is not None:
            assistant_message.metrics["prompt_tokens"] = metrics.prompt_tokens
            self.metrics["prompt_tokens"] = self.metrics.get("prompt_tokens", 0) + metrics.prompt_tokens
        if metrics.completion_tokens is not None:
            assistant_message.metrics["completion_tokens"] = metrics.completion_tokens
            self.metrics["completion_tokens"] = self.metrics.get("completion_tokens", 0) + metrics.completion_tokens
        if metrics.total_tokens is not None:
            assistant_message.metrics["total_tokens"] = metrics.total_tokens
            self.metrics["total_tokens"] = self.metrics.get("total_tokens", 0) + metrics.total_tokens
        if metrics.completion_time is not None:
            assistant_message.metrics["completion_time"] = metrics.completion_time
            self.metrics["completion_time"] = self.metrics.get("completion_time", 0) + metrics.completion_time
        if metrics.prompt_time is not None:
            assistant_message.metrics["prompt_time"] = metrics.prompt_time
            self.metrics["prompt_time"] = self.metrics.get("prompt_time", 0) + metrics.prompt_time
        if metrics.queue_time is not None:
            assistant_message.metrics["queue_time"] = metrics.queue_time
            self.metrics["queue_time"] = self.metrics.get("queue_time", 0) + metrics.queue_time
        if metrics.total_time is not None:
            assistant_message.metrics["total_time"] = metrics.total_time
            self.metrics["total_time"] = self.metrics.get("total_time", 0) + metrics.total_time

    def add_response_usage_to_metrics(self, metrics: Metrics, response_usage: CompletionUsage):
        metrics.input_tokens = response_usage.prompt_tokens
        metrics.prompt_tokens = response_usage.prompt_tokens
        metrics.output_tokens = response_usage.completion_tokens
        metrics.completion_tokens = response_usage.completion_tokens
        metrics.total_tokens = response_usage.total_tokens
        metrics.completion_time = response_usage.completion_time
        metrics.prompt_time = response_usage.prompt_time
        metrics.queue_time = response_usage.queue_time
        metrics.total_time = response_usage.total_time

    def handle_stream_tool_calls(
            self,
            assistant_message: Message,
            messages: List[Message],
            tool_role: str = "tool",
    ) -> Iterator[ModelResponse]:
        """
        Handle tool calls for response stream.

        Args:
            assistant_message (Message): The assistant message.
            messages (List[Message]): The list of messages.
            tool_role (str): The role of the tool call. Defaults to "tool".

        Returns:
            Iterator[ModelResponse]: An iterator of the model response.
        """
        if assistant_message.tool_calls is not None and len(assistant_message.tool_calls) > 0 and self.run_tools:
            function_calls_to_run: List[FunctionCall] = []
            function_call_results: List[Message] = []
            for tool_call in assistant_message.tool_calls:
                _tool_call_id = tool_call.get("id")
                _function_call = get_function_call_for_tool_call(tool_call, self.functions)
                if _function_call is None:
                    messages.append(
                        Message(
                            role=tool_role,
                            tool_call_id=_tool_call_id,
                            content="Could not find function to call.",
                        )
                    )
                    continue
                if _function_call.error is not None:
                    messages.append(
                        Message(
                            role=tool_role,
                            tool_call_id=_tool_call_id,
                            content=_function_call.error,
                        )
                    )
                    continue
                function_calls_to_run.append(_function_call)

            if self.show_tool_calls:
                yield ModelResponse(content="\nRunning:")
                for _f in function_calls_to_run:
                    yield ModelResponse(content=f"\n - {_f.get_call_str()}")
                yield ModelResponse(content="\n\n")

            for function_call_response in self.run_function_calls(
                    function_calls=function_calls_to_run, function_call_results=function_call_results,
                    tool_role=tool_role
            ):
                yield function_call_response

            if len(function_call_results) > 0:
                messages.extend(function_call_results)

    def response_stream(self, messages: List[Message]) -> Iterator[ModelResponse]:
        """
        Generate a streaming response from Groq.

        Args:
            messages (List[Message]): A list of messages.

        Returns:
            Iterator[ModelResponse]: An iterator of model responses.
        """
        self._log_messages(messages)
        stream_data: StreamData = StreamData()
        metrics: Metrics = Metrics()

        # -*- Generate response
        metrics.response_timer.start()
        for response in self.invoke_stream(messages=messages):
            if len(response.choices) > 0:
                metrics.completion_tokens += 1
                if metrics.completion_tokens == 1:
                    metrics.time_to_first_token = metrics.response_timer.elapsed

                response_delta: ChoiceDelta = response.choices[0].delta
                response_content: Optional[str] = response_delta.content
                response_tool_calls: Optional[List[ChoiceDeltaToolCall]] = response_delta.tool_calls

                if response_content is not None:
                    stream_data.response_content += response_content
                    yield ModelResponse(content=response_content)

                if response_tool_calls is not None:
                    if stream_data.response_tool_calls is None:
                        stream_data.response_tool_calls = []
                    stream_data.response_tool_calls.extend(response_tool_calls)

            if response.usage is not None:
                self.add_response_usage_to_metrics(metrics=metrics, response_usage=response.usage)
        metrics.response_timer.stop()

        # -*- Create assistant message
        assistant_message = Message(role="assistant")
        if stream_data.response_content != "":
            assistant_message.content = stream_data.response_content

        if stream_data.response_tool_calls is not None:
            _tool_calls = self.build_tool_calls(stream_data.response_tool_calls)
            if len(_tool_calls) > 0:
                assistant_message.tool_calls = _tool_calls

        # -*- Update usage metrics
        self.update_stream_metrics(assistant_message=assistant_message, metrics=metrics)

        # -*- Add assistant message to messages
        messages.append(assistant_message)

        # -*- Log response and metrics
        assistant_message.log()
        metrics.log()

        # -*- Handle tool calls
        if assistant_message.tool_calls is not None and len(assistant_message.tool_calls) > 0 and self.run_tools:
            tool_role = "tool"
            yield from self.handle_stream_tool_calls(
                assistant_message=assistant_message, messages=messages, tool_role=tool_role
            )
            yield from self.handle_post_tool_call_messages_stream(messages=messages)

    async def aresponse_stream(self, messages: List[Message]) -> Any:
        """
        Generate an asynchronous streaming response from Groq.

        Args:
            messages (List[Message]): A list of messages.

        Returns:
            Any: An asynchronous iterator of model responses.
        """
        self._log_messages(messages)
        stream_data: StreamData = StreamData()
        metrics: Metrics = Metrics()

        # -*- Generate response
        metrics.response_timer.start()
        async for response in self.ainvoke_stream(messages=messages):
            if len(response.choices) > 0:
                metrics.completion_tokens += 1
                if metrics.completion_tokens == 1:
                    metrics.time_to_first_token = metrics.response_timer.elapsed

                response_delta: ChoiceDelta = response.choices[0].delta
                response_content = response_delta.content
                response_tool_calls = response_delta.tool_calls

                if response_content is not None:
                    stream_data.response_content += response_content
                    yield ModelResponse(content=response_content)

                if response_tool_calls is not None:
                    if stream_data.response_tool_calls is None:
                        stream_data.response_tool_calls = []
                    stream_data.response_tool_calls.extend(response_tool_calls)

            if response.usage is not None:
                self.add_response_usage_to_metrics(metrics=metrics, response_usage=response.usage)
        metrics.response_timer.stop()

        # -*- Create assistant message
        assistant_message = Message(role="assistant")
        if stream_data.response_content != "":
            assistant_message.content = stream_data.response_content

        if stream_data.response_tool_calls is not None:
            _tool_calls = self.build_tool_calls(stream_data.response_tool_calls)
            if len(_tool_calls) > 0:
                assistant_message.tool_calls = _tool_calls

        self.update_stream_metrics(assistant_message=assistant_message, metrics=metrics)

        # -*- Add assistant message to messages
        messages.append(assistant_message)

        # -*- Log response and metrics
        assistant_message.log()
        metrics.log()

        # -*- Handle tool calls
        if assistant_message.tool_calls is not None and len(assistant_message.tool_calls) > 0 and self.run_tools:
            tool_role = "tool"
            for tool_call_response in self.handle_stream_tool_calls(
                    assistant_message=assistant_message, messages=messages, tool_role=tool_role
            ):
                yield tool_call_response
            async for post_tool_call_response in self.ahandle_post_tool_call_messages_stream(messages=messages):
                yield post_tool_call_response

    def build_tool_calls(self, tool_calls_data: List[ChoiceDeltaToolCall]) -> List[Dict[str, Any]]:
        """
        Build tool calls from tool call data.

        Args:
            tool_calls_data (List[ChoiceDeltaToolCall]): The tool call data to build from.

        Returns:
            List[Dict[str, Any]]: The built tool calls.
        """
        tool_calls: List[Dict[str, Any]] = []
        for _tool_call in tool_calls_data:
            _index = _tool_call.index
            _tool_call_id = _tool_call.id
            _tool_call_type = _tool_call.type
            _function_name = _tool_call.function.name if _tool_call.function else None
            _function_arguments = _tool_call.function.arguments if _tool_call.function else None

            if len(tool_calls) <= _index:
                tool_calls.extend([{}] * (_index - len(tool_calls) + 1))
            tool_call_entry = tool_calls[_index]
            if not tool_call_entry:
                tool_call_entry["id"] = _tool_call_id
                tool_call_entry["type"] = _tool_call_type
                tool_call_entry["function"] = {
                    "name": _function_name or "",
                    "arguments": _function_arguments or "",
                }
            else:
                if _function_name:
                    tool_call_entry["function"]["name"] += _function_name
                if _function_arguments:
                    tool_call_entry["function"]["arguments"] += _function_arguments
                if _tool_call_id:
                    tool_call_entry["id"] = _tool_call_id
                if _tool_call_type:
                    tool_call_entry["type"] = _tool_call_type
        return tool_calls



from time import time
from enum import Enum
from typing import Optional, Any, Dict

from dataclasses import dataclass


class ModelResponseEvent(str, Enum):
    """Events that can be sent by the Model.response() method"""

    tool_call_started = "ToolCallStarted"
    tool_call_completed = "ToolCallCompleted"
    assistant_response = "AssistantResponse"


@dataclass
class ModelResponse:
    """Response returned by Model.response()"""

    content: Optional[str] = None
    parsed: Optional[Any] = None
    audio: Optional[Dict[str, Any]] = None
    tool_call: Optional[Dict[str, Any]] = None
    event: str = ModelResponseEvent.assistant_response.value
    reasoning_content: Optional[str] = None
    created_at: int = int(time())


class FileType(str, Enum):
    MP4 = "mp4"
    GIF = "gif"
    MP3 = "mp3"



# -*- encoding: utf-8 -*-
"""
@author: orange-crow, XuMing(xuming624@qq.com)
@description:
part of the code is from phidata
"""
import collections.abc
import io
import base64
from types import GeneratorType
from typing import List, Iterator, Optional, Dict, Any, Callable, Union, Sequence

from pydantic import BaseModel, ConfigDict, Field, field_validator, ValidationInfo

from agentica.model.message import Message
from agentica.model.response import ModelResponse, ModelResponseEvent
from agentica.tools.base import ModelTool, Tool, Function, FunctionCall, ToolCallException
from agentica.utils.log import logger
from agentica.utils.timer import Timer


class Model(BaseModel):
    # ID of the model to use.
    id: str = Field(..., alias="model")
    # Name for this Model. This is not sent to the Model API.
    name: Optional[str] = None
    # Provider for this Model. This is not sent to the Model API.
    provider: Optional[str] = Field(None, validate_default=True)
    # Metrics collected for this Model. This is not sent to the Model API.
    metrics: Dict[str, Any] = Field(default_factory=dict)
    response_format: Optional[Any] = None

    # A list of tools provided to the Model.
    # Tools are functions the model may generate JSON inputs for.
    # If you provide a dict, it is not called by the model.
    # Always add tools using the add_tool() method.
    tools: Optional[List[Union[ModelTool, Dict]]] = None
    # Controls which (if any) function is called by the model.
    # "none" means the model will not call a function and instead generates a message.
    # "auto" means the model can pick between generating a message or calling a function.
    # Specifying a particular function via {"type: "function", "function": {"name": "my_function"}}
    #   forces the model to call that function.
    # "none" is the default when no functions are present. "auto" is the default if functions are present.
    tool_choice: Optional[Union[str, Dict[str, Any]]] = None
    # If True, runs the tool before sending back the response content.
    run_tools: bool = True
    # If True, shows function calls in the response.
    show_tool_calls: Optional[bool] = None
    # Maximum number of tool calls allowed.
    tool_call_limit: Optional[int] = None

    # -*- Functions available to the Model to call -*-
    # Functions extracted from the tools.
    # Note: These are not sent to the Model API and are only used for execution + deduplication.
    functions: Optional[Dict[str, Function]] = None
    # Function call stack.
    function_call_stack: Optional[List[FunctionCall]] = None

    # System prompt from the model added to the Agent.
    system_prompt: Optional[str] = None
    # Instructions from the model added to the Agent.
    instructions: Optional[List[str]] = None

    # Session ID of the calling Agent or Workflow.
    session_id: Optional[str] = None
    # Whether to use the structured outputs with this Model.
    structured_outputs: Optional[bool] = None
    # Whether the Model supports structured outputs.
    supports_structured_outputs: bool = False

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    @field_validator("provider", mode="before")
    def set_provider(cls, v: Optional[str], info: ValidationInfo) -> str:
        model_name = info.data.get("name")
        model_id = info.data.get("id")
        return v or f"{model_name} ({model_id})"

    @property
    def request_kwargs(self) -> Dict[str, Any]:
        raise NotImplementedError

    def to_dict(self) -> Dict[str, Any]:
        _dict = self.model_dump(include={"name", "id", "provider", "metrics"})
        if self.functions:
            _dict["functions"] = {k: v.to_dict() for k, v in self.functions.items()}
            _dict["tool_call_limit"] = self.tool_call_limit
        return _dict

    def invoke(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    async def ainvoke(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def invoke_stream(self, *args, **kwargs) -> Iterator[Any]:
        raise NotImplementedError

    async def ainvoke_stream(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def response(self, messages: List[Message]) -> ModelResponse:
        raise NotImplementedError

    async def aresponse(self, messages: List[Message]) -> ModelResponse:
        raise NotImplementedError

    def response_stream(self, messages: List[Message]) -> Iterator[ModelResponse]:
        raise NotImplementedError

    async def aresponse_stream(self, messages: List[Message]) -> Any:
        raise NotImplementedError

    def _log_messages(self, messages: List[Message]) -> None:
        """
        Log messages for debugging.
        """
        for m in messages:
            m.log()

    def get_tools_for_api(self) -> Optional[List[Dict[str, Any]]]:
        if self.tools is None:
            return None

        tools_for_api = []
        for tool in self.tools:
            if isinstance(tool, ModelTool):
                tools_for_api.append(tool.to_dict())
            elif isinstance(tool, Dict):
                tools_for_api.append(tool)
        return tools_for_api

    def add_tool(
            self, tool: Union[ModelTool, Tool, Callable, Dict, Function], strict: bool = False,
            agent: Optional[Any] = None
    ) -> None:
        if self.tools is None:
            self.tools = []

        # If the tool is a Tool or Dict, add it directly to the Model
        if isinstance(tool, ModelTool) or isinstance(tool, Dict):
            if tool not in self.tools:
                self.tools.append(tool)
                logger.debug(f"Added tool {tool} to model.")

        # If the tool is a Callable or Toolkit, process and add to the Model
        elif callable(tool) or isinstance(tool, Tool) or isinstance(tool, Function):
            if self.functions is None:
                self.functions = {}

            if isinstance(tool, Tool):
                # For each function in the toolkit, process entrypoint and add to self.tools
                for name, func in tool.functions.items():
                    # If the function does not exist in self.functions, add to self.tools
                    if name not in self.functions:
                        func._agent = agent
                        func.process_entrypoint(strict=strict)
                        if strict and self.supports_structured_outputs:
                            func.strict = True
                        self.functions[name] = func
                        self.tools.append({"type": "function", "function": func.to_dict()})
                        logger.debug(f"Function {name} from {tool.name} added to model.")

            elif isinstance(tool, Function):
                if tool.name not in self.functions:
                    tool._agent = agent
                    tool.process_entrypoint(strict=strict)
                    if strict and self.supports_structured_outputs:
                        tool.strict = True
                    self.functions[tool.name] = tool
                    self.tools.append({"type": "function", "function": tool.to_dict()})
                    logger.debug(f"Function {tool.name} added to model.")

            elif callable(tool):
                try:
                    function_name = tool.__name__
                    if function_name not in self.functions:
                        func = Function.from_callable(tool, strict=strict)
                        func._agent = agent
                        if strict and self.supports_structured_outputs:
                            func.strict = True
                        self.functions[func.name] = func
                        self.tools.append({"type": "function", "function": func.to_dict()})
                        logger.debug(f"Function {func.name} added to model.")
                except Exception as e:
                    logger.warning(f"Could not add function {tool}: {e}")

    def deactivate_function_calls(self) -> None:
        # Deactivate tool calls by setting future tool calls to "none"
        # This is triggered when the function call limit is reached.
        self.tool_choice = "none"

    def run_function_calls(
            self, function_calls: List[FunctionCall], function_call_results: List[Message], tool_role: str = "tool"
    ) -> Iterator[ModelResponse]:
        for function_call in function_calls:
            if self.function_call_stack is None:
                self.function_call_stack = []

            # -*- Start function call
            function_call_timer = Timer()
            function_call_timer.start()
            yield ModelResponse(
                content=function_call.get_call_str(),
                tool_call={
                    "role": tool_role,
                    "tool_call_id": function_call.call_id,
                    "tool_name": function_call.function.name,
                    "tool_args": function_call.arguments,
                },
                event=ModelResponseEvent.tool_call_started.value,
            )

            # Track if the function call was successful
            function_call_success = False
            # If True, stop execution after this function call
            stop_execution_after_tool_call = False
            # Additional messages from the function call that will be added to the function call results
            additional_messages_from_function_call = []

            # -*- Run function call
            try:
                function_call_success = function_call.execute()
            except ToolCallException as tce:
                if tce.user_message is not None:
                    if isinstance(tce.user_message, str):
                        additional_messages_from_function_call.append(Message(role="user", content=tce.user_message))
                    else:
                        additional_messages_from_function_call.append(tce.user_message)
                if tce.agent_message is not None:
                    if isinstance(tce.agent_message, str):
                        additional_messages_from_function_call.append(
                            Message(role="assistant", content=tce.agent_message)
                        )
                    else:
                        additional_messages_from_function_call.append(tce.agent_message)
                if tce.messages is not None and len(tce.messages) > 0:
                    for m in tce.messages:
                        if isinstance(m, Message):
                            additional_messages_from_function_call.append(m)
                        elif isinstance(m, dict):
                            try:
                                additional_messages_from_function_call.append(Message(**m))
                            except Exception as e:
                                logger.warning(f"Failed to convert dict to Message: {e}")
                if tce.stop_execution:
                    stop_execution_after_tool_call = True
                    if len(additional_messages_from_function_call) > 0:
                        for m in additional_messages_from_function_call:
                            m.stop_after_tool_call = True

            function_call_output: Optional[Union[List[Any], str]] = ""
            if isinstance(function_call.result, (GeneratorType, collections.abc.Iterator)):
                for item in function_call.result:
                    function_call_output += item
                    if function_call.function.show_result:
                        yield ModelResponse(content=item)
            else:
                function_call_output = function_call.result
                if function_call.function.show_result:
                    yield ModelResponse(content=function_call_output)

            # -*- Stop function call timer
            function_call_timer.stop()

            # -*- Create function call result message
            function_call_result = Message(
                role=tool_role,
                content=function_call_output if function_call_success else function_call.error,
                tool_call_id=function_call.call_id,
                tool_name=function_call.function.name,
                tool_args=function_call.arguments,
                tool_call_error=not function_call_success,
                stop_after_tool_call=function_call.function.stop_after_tool_call or stop_execution_after_tool_call,
                metrics={"time": function_call_timer.elapsed},
            )

            # -*- Yield function call result
            yield ModelResponse(
                content=f"{function_call.get_call_str()} completed in {function_call_timer.elapsed:.4f}s.",
                tool_call=function_call_result.model_dump(
                    include={
                        "content",
                        "tool_call_id",
                        "tool_name",
                        "tool_args",
                        "tool_call_error",
                        "metrics",
                        "created_at",
                    }
                ),
                event=ModelResponseEvent.tool_call_completed.value,
            )

            # Add metrics to the model
            if "tool_call_times" not in self.metrics:
                self.metrics["tool_call_times"] = {}
            if function_call.function.name not in self.metrics["tool_call_times"]:
                self.metrics["tool_call_times"][function_call.function.name] = []
            self.metrics["tool_call_times"][function_call.function.name].append(function_call_timer.elapsed)

            # Add the function call result to the function call results
            function_call_results.append(function_call_result)
            if len(additional_messages_from_function_call) > 0:
                function_call_results.extend(additional_messages_from_function_call)
            self.function_call_stack.append(function_call)

            # -*- Check function call limit
            if self.tool_call_limit and len(self.function_call_stack) >= self.tool_call_limit:
                self.deactivate_function_calls()
                break  # Exit early if we reach the function call limit

    def handle_post_tool_call_messages(self, messages: List[Message], model_response: ModelResponse) -> ModelResponse:
        last_message = messages[-1]
        if last_message.stop_after_tool_call:
            logger.debug("Stopping execution as stop_after_tool_call=True")
            if (
                    last_message.role == "assistant"
                    and last_message.content is not None
                    and isinstance(last_message.content, str)
            ):
                if model_response.content is None:
                    model_response.content = ""
                model_response.content += last_message.content
        else:
            response_after_tool_calls = self.response(messages=messages)
            if response_after_tool_calls.content is not None:
                if model_response.content is None:
                    model_response.content = ""
                model_response.content += response_after_tool_calls.content
            if response_after_tool_calls.parsed is not None:
                # bubble up the parsed object, so that the final response has the parsed object
                # that is visible to the agent
                model_response.parsed = response_after_tool_calls.parsed
            if response_after_tool_calls.audio is not None:
                # bubble up the audio, so that the final response has the audio
                # that is visible to the agent
                model_response.audio = response_after_tool_calls.audio
        return model_response

    async def ahandle_post_tool_call_messages(
            self, messages: List[Message], model_response: ModelResponse
    ) -> ModelResponse:
        last_message = messages[-1]
        if last_message.stop_after_tool_call:
            logger.debug("Stopping execution as stop_after_tool_call=True")
            if (
                    last_message.role == "assistant"
                    and last_message.content is not None
                    and isinstance(last_message.content, str)
            ):
                if model_response.content is None:
                    model_response.content = ""
                model_response.content += last_message.content
        else:
            response_after_tool_calls = await self.aresponse(messages=messages)
            if response_after_tool_calls.content is not None:
                if model_response.content is None:
                    model_response.content = ""
                model_response.content += response_after_tool_calls.content
            if response_after_tool_calls.parsed is not None:
                # bubble up the parsed object, so that the final response has the parsed object
                # that is visible to the agent
                model_response.parsed = response_after_tool_calls.parsed
            if response_after_tool_calls.audio is not None:
                # bubble up the audio, so that the final response has the audio
                # that is visible to the agent
                model_response.audio = response_after_tool_calls.audio
        return model_response

    def handle_post_tool_call_messages_stream(self, messages: List[Message]) -> Iterator[ModelResponse]:
        last_message = messages[-1]
        if last_message.stop_after_tool_call:
            logger.debug("Stopping execution as stop_after_tool_call=True")
            if (
                    last_message.role == "assistant"
                    and last_message.content is not None
                    and isinstance(last_message.content, str)
            ):
                yield ModelResponse(content=last_message.content)
        else:
            yield from self.response_stream(messages=messages)

    async def ahandle_post_tool_call_messages_stream(self, messages: List[Message]) -> Any:
        last_message = messages[-1]
        if last_message.stop_after_tool_call:
            logger.debug("Stopping execution as stop_after_tool_call=True")
            if (
                    last_message.role == "assistant"
                    and last_message.content is not None
                    and isinstance(last_message.content, str)
            ):
                yield ModelResponse(content=last_message.content)
        else:
            async for model_response in self.aresponse_stream(messages=messages):  # type: ignore
                yield model_response

    def _process_string_image(self, image: str) -> Dict[str, Any]:
        """Process string-based image (base64, URL, or file path)."""

        # Process Base64 encoded image
        if image.startswith("data:image"):
            return {"type": "image_url", "image_url": {"url": image}}

        # Process URL image
        if image.startswith(("http://", "https://")):
            return {"type": "image_url", "image_url": {"url": image}}

        # Process local file image
        import mimetypes
        from pathlib import Path

        path = Path(image)
        if not path.exists():
            raise FileNotFoundError(f"Image file not found: {image}")

        mime_type = mimetypes.guess_type(image)[0] or "image/jpeg"
        with open(path, "rb") as image_file:
            base64_image = base64.b64encode(image_file.read()).decode("utf-8")
            image_url = f"data:{mime_type};base64,{base64_image}"
            return {"type": "image_url", "image_url": {"url": image_url}}

    def _process_pil_image(self, image: 'PIL.Image.Image') -> Dict[str, Any]:
        """Process PIL Image data."""
        # Convert image to bytes
        img_byte_arr = io.BytesIO()
        image.save(img_byte_arr, format='PNG')
        img_byte_arr = img_byte_arr.getvalue()

        # Convert to base64
        base64_image = base64.b64encode(img_byte_arr).decode('utf-8')
        image_url = f"data:image/png;base64,{base64_image}"
        return {"type": "image_url", "image_url": {"url": image_url}}

    def _process_bytes_image(self, image: bytes) -> Dict[str, Any]:
        """Process bytes image data."""
        base64_image = base64.b64encode(image).decode("utf-8")
        image_url = f"data:image/jpeg;base64,{base64_image}"
        return {"type": "image_url", "image_url": {"url": image_url}}

    def process_image(self, image: Any) -> Optional[Dict[str, Any]]:
        """Process an image based on the format."""
        from PIL.Image import Image as PILImage
        if isinstance(image, dict):
            return {"type": "image_url", "image_url": image}

        if isinstance(image, str):
            return self._process_string_image(image)

        if isinstance(image, bytes):
            return self._process_bytes_image(image)

        if isinstance(image, PILImage):
            return self._process_pil_image(image)

        logger.warning(f"Unsupported image type: {type(image)}")
        return None

    def add_images_to_message(self, message: Message, images: Optional[Sequence[Any]] = None) -> Message:
        """
        Add images to a message for the model. By default, we use the OpenAI image format but other Models
        can override this method to use a different image format.
        Args:
            message: The message for the Model
            images: Sequence of images in various formats:
                - str: base64 encoded image, URL, or file path
                - Dict: pre-formatted image data
                - bytes: raw image data

        Returns:
            Message content with images added in the format expected by the model
        """
        # If no images are provided, return the message as is
        if images is None or len(images) == 0:
            return message

        # Ignore non-string message content
        # because we assume that the images/audio are already added to the message
        if not isinstance(message.content, str):
            return message

        # Create a default message content with text
        message_content_with_image: List[Dict[str, Any]] = [{"type": "text", "text": message.content}]

        # Add images to the message content
        for image in images:
            try:
                image_data = self.process_image(image)
                if image_data:
                    message_content_with_image.append(image_data)
            except Exception as e:
                logger.error(f"Failed to process image: {str(e)}")
                continue

        # Update the message content with the images
        message.content = message_content_with_image
        return message

    def add_audio_to_message(self, message: Message, audio: Optional[Any] = None) -> Message:
        """
        Add audio to a message for the model. By default, we use the OpenAI audio format but other Models
        can override this method to use a different audio format.
        Args:
            message: The message for the Model
            audio: Pre-formatted audio data like {
                        "data": encoded_string,
                        "format": "wav"
                    }

        Returns:
            Message content with audio added in the format expected by the model
        """
        if audio is None:
            return message

        # If `id` is in the audio, this means the audio is already processed
        # This is used in multi-turn conversations
        if "id" in audio:
            message.content = ""
            message.audio = {"id": audio["id"]}
        # If `data` is in the audio, this means the audio is raw data
        # And an input audio
        elif "data" in audio:
            # Create a message with audio
            message.content = [
                {"type": "text", "text": message.content},
                {"type": "input_audio", "input_audio": audio},
            ]
        return message

    def get_system_message_for_model(self) -> Optional[str]:
        return self.system_prompt

    def get_instructions_for_model(self) -> Optional[List[str]]:
        return self.instructions

    def clear(self) -> None:
        """Clears the Model's state."""

        self.metrics = {}
        self.functions = None
        self.function_call_stack = None
        self.session_id = None

    def deep_copy(self, *, update: Optional[Dict[str, Any]] = None) -> "Model":
        new_model = self.model_copy(deep=True, update=update)
        # Clear the new model to remove any references to the old model
        new_model.clear()
        return new_model



from typing import Optional, Any

from pydantic import BaseModel, model_validator

__all__ = ["Media", "Video", "Image", "Audio"]


class Media(BaseModel):
    id: str
    original_prompt: Optional[str] = None
    revised_prompt: Optional[str] = None


class Video(Media):
    url: str  # Remote location for file
    eta: Optional[str] = None
    length: Optional[str] = None


class Image(Media):
    url: str  # Remote location for file
    alt_text: Optional[str] = None


class Audio(Media):
    url: Optional[str] = None  # Remote location for file
    base64_audio: Optional[str] = None  # Base64-encoded audio data
    length: Optional[str] = None

    @model_validator(mode="before")
    def validate_exclusive_audio(cls, data: Any):
        """
        Ensure that either `url` or `base64_audio` is provided, but not both.
        """
        if data.get("url") and data.get("base64_audio"):
            raise ValueError("Provide either `url` or `base64_audio`, not both.")
        if not data.get("url") and not data.get("base64_audio"):
            raise ValueError("Either `url` or `base64_audio` must be provided.")
        return data



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
"""
from typing import Optional

from agentica.model.message import Message
from agentica.model.base import Model
from agentica.model.openai import OpenAIChat
from agentica.tools.base import Tool


class TextAnalysisTool(Tool):
    def __init__(self, llm: Optional[Model] = None):
        super().__init__(name="text_analysis_tool")
        self.llm = llm
        self.register(self.text_analysis_use_llm)

    def update_llm(self) -> None:
        if self.llm is None:
            self.llm = OpenAIChat()

    def text_analysis_use_llm(self, prompt: str) -> str:
        """Use LLM to analyze text, including sentiment analysis, entity recognition, text summarization,
        and information extraction, etc.

        Args:
            prompt (str): The prompt to analyze, send to LLM.

        Example:
            ```python
            from agentica.tools.text_analysis_tool import TextAnalysisTool
            tool = TextAnalysisTool()
            result = tool.text_analysis_use_llm("分析下面内容的情感：这面条好难吃啊。")
            print(result)
            ```

        Returns:
            str: The result of the analysis.
        """
        self.update_llm()
        llm_messages = [Message.model_validate({"role": "user", "content": prompt})]
        return self.llm.response(llm_messages).content


if __name__ == '__main__':
    # from agentica.tools.text_analysis_tool import TextAnalysisTool
    tool = TextAnalysisTool()
    result = tool.text_analysis_use_llm("分析下面内容的情感：这面条好难吃啊。")
    print(result)



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
from __future__ import annotations

from typing import Optional, Dict, Any

from pydantic import BaseModel, ConfigDict


class AgentSession(BaseModel):
    """Agent Session that is stored in the database"""

    # Session UUID
    session_id: str
    # ID of the agent that this session is associated with
    agent_id: Optional[str] = None
    # ID of the user interacting with this agent
    user_id: Optional[str] = None
    # Agent Memory
    memory: Optional[Dict[str, Any]] = None
    # Agent Metadata
    agent_data: Optional[Dict[str, Any]] = None
    # User Metadata
    user_data: Optional[Dict[str, Any]] = None
    # Session Metadata
    session_data: Optional[Dict[str, Any]] = None
    # The Unix timestamp when this session was created
    created_at: Optional[int] = None
    # The Unix timestamp when this session was last updated
    updated_at: Optional[int] = None

    model_config = ConfigDict(from_attributes=True)

    def monitoring_data(self) -> Dict[str, Any]:
        # Google Gemini adds a "parts" field to the messages, which is not serializable
        # If the provider is Google, remove the "parts" from the messages
        if self.agent_data is not None:
            if self.agent_data.get("model", {}).get("provider") == "Google" and self.memory is not None:
                # Remove parts from runs' response messages
                if "runs" in self.memory:
                    for _run in self.memory["runs"]:
                        if "response" in _run and "messages" in _run["response"]:
                            for m in _run["response"]["messages"]:
                                if isinstance(m, dict):
                                    m.pop("parts", None)

                # Remove parts from top-level memory messages
                if "messages" in self.memory:
                    for m in self.memory["messages"]:
                        if isinstance(m, dict):
                            m.pop("parts", None)

        monitoring_data = self.model_dump()
        return monitoring_data

    def telemetry_data(self) -> Dict[str, Any]:
        return self.model_dump(include={"model", "created_at", "updated_at"})
